# -*- coding: utf-8 -*-

# SAX
from xml.sax import ContentHandler
from xml.sax import make_parser
from xml.sax import parseString

# para el cliente http
import httplib

from pandac.PandaModules import Point3

from direct.task.Task import Task

#from pandac.PandaModules import HTTPClient
#from pandac.PandaModules import Ramfile
#from pandac.PandaModules import URLSpec

from escalas import Escalas
from Dijkstra import shortest_path

# define los limites del mapa
class Bounds:
    
    
    # Coordenadas geográficas
    def setMinCoord(self,coord):
        self.mincoord = coord
    def setMaxCoord(self,coord):
        self.maxcoord = coord
    
    def getMinCoord(self):
        return self.mincoord
    def getMaxCoord(self):
        return self.maxcoord
    
    #Coordenadas cartesianas
    def setMinPos(self,p):
        self.minpos = p
    def getMinPos(self):
        return self.minpos

    def setMaxPos(self,p):
        self.maxpos = p
    def getMaxPos(self):
        return self.maxpos

# un Nodo es una posicion concreta del mapa
class Nodo:
    
    def __init__(self):
        
        # diccionario de etiquetas del nodo
        self.Tags = {}
        # lista de way donde el nodo aparece
        self.Ways = []
        
        # indica si esta úicamente en una carretera de un solo carril
        self.oneway = 'yes'
    
        # nombre de way al que pertenece el nodo
        self.wayName = []
        
        # lista de adyacencia de la matriz de camino
        self.listaAdyacencia = None
        
    def __str__(self):
        print self.getId()
        
    def __cmp__(self,otro):
        return int(self.getId()) - int(otro.getId())
        
    def setDatos(self,id,lon,lat):
        self.n = (id,lon,lat)     
        
    def getId(self):
        return self.n[0]
    def getLon(self):
        return self.n[1]
    def getLat(self):
        return self.n[2]
    
    # incluye una nueva tupla (key,value) a la lista de tags
    def addTag(self,k,v):
        self.Tags[k] = v
    def addTags(self,t):
        self.Tags = t
    def getTags(self):
        return self.Tags
    # si way tiene la clave la devuelve, None en caso contrario
    def getTagValue(self,tag):
        return self.Tags.get(tag,None)
    # comprueba si way tiene la clave    
    def has_tag(self,tag):
        return self.Tags.has_key(tag)

    def setPos(self,p):
        self.pos = p
    def getPos(self):
        return self.pos
    def getX(self):
        return self.pos.getX()
    def getY(self):
        return self.pos.getY()
    
    def setOneWay(self,o):
        self.oneway = o
    def getOneWay(self):
        return self.oneway
    
    def addWayName(self,wn):
        if not self.wayName.__contains__(wn):
            self.wayName.append(wn)
        
    def getWayName(self):
        
        if self.wayName.__len__()==0:
            return ''
        else:
            
            str = self.wayName[0]
            for s in self.wayName[1:]:
                str = str + ' con ' + s
            
            return str
    
    """
    Se utiliza en la clase escenario para saber que way pertenecen
    a los nodos dentro la porcion de mapa a pintar
    """
    def addWay(self,w):
        self.Ways.append(w)
    def getListaWays(self):
        return self.Ways

    """
        Por compatibilidad con el metodo getListaRefs
        de la clase Way utilizado indistintamente en el
        metodo CalcularPorcionMapa
    """
    def getListaRefs(self):
        return list([self])
    
    # crea la lista de adyacencia de la matriz de camino
    def CrearListaAdyacencia(self):
        
        if self.listaAdyacencia == None:
            self.listaAdyacencia = []
        
    # añade un elemento a la lista de adyacencia
    def addListaAdyacencia(self,nd):
        self.listaAdyacencia.append(nd)
    
    # devuelve una copia de la lista de adyacencia
    def getListaAdyacentes(self):
        return list(self.listaAdyacencia)
    
    # devuelve una copia de la lista de adyacencia sin nd
    def getListaSiguientes(self,nodo):
        
        listaSiguientes = []
        
        for nd in self.listaAdyacencia:
            if nd != nodo:
                listaSiguientes.append(nd)
        
        return listaSiguientes
    

# Describe cualquier elemento que pueda aparecer en el mapa
class Way:
    
    def __init__(self):
        
        # id de way
        self.id = None
        # lista de ref de way
        self.Refs = []
        # diccionario de tags, {Key:Value}
        self.Tags = {}
        
        # identifica si es una carretera, un parque, una plaza ...
        self.tipo = None

    # establece el valor id
    def setId(self,i):
        self.id = i
    # devuelve id de way
    def getId(self):
        return self.id
    
    def getName(self):
        return self.Tags.get('name','Calle id: '+self.id)
    
    # incluye una nueva referencia a la lista de referencias
    def addRef(self,r):
        self.Refs.append(r)
    # lista de referencias de way
    def getListaRefs(self):
        return list(self.Refs)
    
    # incluye una nueva tupla (key,value) a la lista de tags
    def addTag(self,k,v):
        self.Tags[k] = v 
    def addTags(self,t):
        self.Tags = t 
    # lista de tags de way
    def getTags(self):
        return self.Tags
    
    # si way tiene la clave la devuelve, None en caso contrario
    def getTagValue(self,tag):
        return self.Tags.get(tag,None)
    
    # comprueba si way tiene la clave    
    def has_tag(self,tag):
        return self.Tags.has_key(tag)

# LA MATRIZ DE CAMINO ES INCLUIDA EN EL PROPIO NODO
## Describe la matriz de camino del mapa
#class MatrizCamino():
#    
#    def __init__(self,carreteras):
#        
#        print "Matriz de camino"
#        
#        self.MatrizCamino = {}
#        
#        for w in carreteras:
#            
#            # lista de nodos que contien way
#            refs = w.getListaRefs()
#            nrefs = refs.__len__()
#            
#            # determina si way es de doble sentido
#            if w.getTagValue('oneway') == 'yes' or w.getTagValue('oneway') == '1':
#                self.oneway = 'yes'
#            else:
#                self.oneway = 'no'
#            
#            self.wayName = w.getTagValue('name')
#            if self.wayName == None:
#                if w.getTagValue('ref') != None:
#                    self.wayName = w.getTagValue('ref')
#                else:
#                    self.wayName = '('+w.getId()+')'
#                    
#                    
#            i=0
#            while i < (nrefs-1) :
#                 
#                if not self.MatrizCamino.has_key( refs[i].getId() ):
#                    self.MatrizCamino[ refs[i].getId() ] = []
#                    
#                # incluye en la lista que estan delante al siguiente nodo
#                self.MatrizCamino[ refs[i].getId() ].append( refs[i+1].getId() )
#                
#                # si way es de doble sentido crea el enlace inverso
#                if self.oneway == 'no':
#                    
#                    # si el nodo no esta en el diccionario lo registra
#                    if not self.MatrizCamino.has_key( refs[i+1].getId() ):
#                        self.MatrizCamino[ refs[i+1].getId() ] = []
#                    
#                    self.MatrizCamino[ refs[i+1].getId() ].append( refs[i].getId() )
#                
#                    
#                if self.wayName != None:
#                    refs[i].addWayName(self.wayName)
#                if self.oneway == 'no':
#                    refs[i].setOneWay(self.oneway)
#            
#                
#                i = i+1
#            # end-while
#                
#            if self.oneway == 'yes':
#                if not self.MatrizCamino.has_key( refs[i].getId() ):
#                    self.MatrizCamino[ refs[i].getId() ] = []
#                
#            if self.wayName != None:
#                refs[i].addWayName(self.wayName)
#            if self.oneway == 'no':
#                refs[i].setOneWay(self.oneway)
#    
#    def getMatriz(self):
#        return self.MatrizCamino
#        
#    def getListaNodos(self):
#        return self.MatrizCamino.keys()
#            
#    def getListaAdyacentes(self,id):
#        return self.MatrizCamino[id]
#    
#    def has_nodo(self,nd):
#        return self.MatrizCamino.has_key(nd)
        
# clase que gestiona los eventos produccidos por sax
class Manejador(ContentHandler):
    
    def __init__(self,bounds,nodos,ways,elementosMapa,iconosMapa):
        
        # copia local de bounds
        self.B = bounds
        # copia local del diccionario de nodos
        self.N = nodos
        # copia local de ways, objeto que contiene la lista de ways
        self.W = ways
        # nodo auxiliar para almacenar los datos que contiene
        self.nodo_aux = Nodo()
        #way auxuliar para almacenar los datos de way mientras se termina de leer
        self.way_aux = Way()
        # para almacenar temporalmente los tags
        self.tags_aux = {}
    
        # diccionario que agrupo los elementos de una misa clase en listas
        self.elementosMapa = elementosMapa
        
        # lista para las carreteras en general
        self.elementosMapa['carreteras'] = []
        # lista para las calles peatonales
        self.elementosMapa['peatonales'] = []
        # lista para las plazas
        self.elementosMapa['plazas'] = []
        # lista para los jardines
        self.elementosMapa['garden'] = []
        
        # lista de iconos
        self.elementosMapa['icono_restaurante'] = []
        self.elementosMapa['icono_bar'] = []
        self.elementosMapa['icono_cafe'] = []
        self.elementosMapa['icono_guarderia'] = []
        self.elementosMapa['icono_colegio'] = []
        self.elementosMapa['icono_libreria'] = []
        self.elementosMapa['icono_universidad'] = []
        self.elementosMapa['icono_autobus'] = []
        self.elementosMapa['icono_gasolinera'] = []
        self.elementosMapa['icono_parking'] = []
        self.elementosMapa['icono_taxi'] = []
        self.elementosMapa['icono_cajero'] = []
        self.elementosMapa['icono_banco'] = []
        self.elementosMapa['icono_farmacia'] = []
        self.elementosMapa['icono_hospital'] = []
        self.elementosMapa['icono_cine'] = []
        self.elementosMapa['icono_teatro'] = []
        self.elementosMapa['icono_iglesia'] = []
        self.elementosMapa['icono_mezquita'] = []
        self.elementosMapa['icono_policia'] = []
        self.elementosMapa['icono_oficina_correos'] = []
        self.elementosMapa['icono_museo'] = []
        self.elementosMapa['icono_hotel'] = []
  
        # lista de areas
        self.elementosMapa['area_restaurante'] = []
        self.elementosMapa['area_bar'] = []
        self.elementosMapa['area_cafe'] = []
        self.elementosMapa['area_guarderia'] = []
        self.elementosMapa['area_colegio'] = []
        self.elementosMapa['area_libreria'] = []
        self.elementosMapa['area_universidad'] = []
        self.elementosMapa['area_autobus'] = []
        self.elementosMapa['area_gasolinera'] = []
        self.elementosMapa['area_parking'] = []
        self.elementosMapa['area_taxi'] = []
        self.elementosMapa['area_cajero'] = []
        self.elementosMapa['area_banco'] = []
        self.elementosMapa['area_farmacia'] = []
        self.elementosMapa['area_hospital'] = []
        self.elementosMapa['area_cine'] = []
        self.elementosMapa['area_teatro'] = []
        self.elementosMapa['area_iglesia'] = []
        self.elementosMapa['area_mezquita'] = []
        self.elementosMapa['area_policia'] = []
        self.elementosMapa['area_oficina_correos'] = []
        self.elementosMapa['area_museo'] = []
        self.elementosMapa['area_hotel'] = []
        
        #edificios
        self.elementosMapa['edificios'] = []
  
    # se ejecuta la comiezo de cada elemento
    def startElement(self,nombre,atributos):
    
        if nombre == 'bounds':

            self.B.setMinCoord(Point3(float(atributos.get('minlon')),
                                      float(atributos.get('minlat')),
                                      0.0))
            
            self.B.setMinPos(Point3(0.0,0.0,0.0))
            
            self.B.setMaxCoord(Point3(float(atributos.get('maxlon')),
                                      float(atributos.get('maxlat')),
                                      0.0))
            
            self.B.setMaxPos( self.B.getMaxCoord()*Escalas['gr-px'] )
        
        if nombre == 'node':
            self.nodo_aux.setDatos(atributos.get('id').encode('utf-8'), 
                                   float(atributos.get('lon')),
                                   float(atributos.get('lat')))
        
            self.nodo_aux.setPos((Point3(self.nodo_aux.getLon(),
                                         self.nodo_aux.getLat(),
                                         0.0) - self.B.getMinCoord()) * Escalas['gr-px'])            
            
        if nombre == 'way':
            # crea un objeto way auxiliar con su id
            self.way_aux.setId( atributos.get('id').encode('utf-8') )
            
        if nombre == 'nd':
            # incluye una nueva referencia al way auxiliar 
            self.way_aux.addRef(self.N[atributos.get('ref')])
            
            self.N[atributos.get('ref')].addWay(self.way_aux)
        
        if nombre == 'tag':
            # encode() transforma a utf-8 y evita problemas con los caracteres
            self.tags_aux[atributos.get('k')] = atributos.get('v').encode('utf-8')                                  
                                 
    def endElement(self,nombre):
        
        if nombre == 'node':
            
            self.nodo_aux.addTags(self.tags_aux)
            self.tags_aux = {}
            
            if self.nodo_aux.getTagValue('amenity') == 'restaurant':
                self.elementosMapa['icono_restaurante'].append(self.nodo_aux)
                
            elif self.nodo_aux.getTagValue('amenity') == 'food_court':
                self.elementosMapa['icono_restaurante'].append(self.nodo_aux)
            
            elif self.nodo_aux.getTagValue('amenity') == 'fast_food':
                self.elementosMapa['icono_restaurante'].append(self.nodo_aux)
            
            elif self.nodo_aux.getTagValue('amenity') == 'pub':
                self.elementosMapa['icono_bar'].append(self.nodo_aux)
            
            elif self.nodo_aux.getTagValue('amenity') == 'bar':
                self.elementosMapa['icono_bar'].append(self.nodo_aux)
                
            elif self.nodo_aux.getTagValue('amenity') == 'cafe':
                self.elementosMapa['icono_cafe'].append(self.nodo_aux)
                
            elif self.nodo_aux.getTagValue('amenity') == 'kindergarten':
                self.elementosMapa['icono_guarderia'].append(self.nodo_aux)
            
            elif self.nodo_aux.getTagValue('amenity') == 'school':
                self.elementosMapa['icono_colegio'].append(self.nodo_aux)
            
            elif self.nodo_aux.getTagValue('amenity') == 'college':
                self.elementosMapa['icono_colegio'].append(self.nodo_aux)
            
            elif self.nodo_aux.getTagValue('amenity') == 'library':
                self.elementosMapa['icono_libreria'].append(self.nodo_aux)
            
            elif self.nodo_aux.getTagValue('amenity') == 'university':
                self.elementosMapa['icono_universidad'].append(self.nodo_aux)
            
            elif self.nodo_aux.getTagValue('amenity') == 'bus_station':
                self.elementosMapa['icono_autobus'].append(self.nodo_aux)
            
            elif self.nodo_aux.getTagValue('highway') == 'bus_stop':
                self.elementosMapa['icono_autobus'].append(self.nodo_aux)
            
            elif self.nodo_aux.getTagValue('amenity') == 'fuel':
                self.elementosMapa['icono_gasolinera'].append(self.nodo_aux)
            
            elif self.nodo_aux.getTagValue('amenity') == 'parking':
                self.elementosMapa['icono_parking'].append(self.nodo_aux)
                
            elif self.nodo_aux.getTagValue('amenity') == 'taxi':
                self.elementosMapa['icono_taxi'].append(self.nodo_aux)
            
            elif self.nodo_aux.getTagValue('amenity') == 'atm':
                self.elementosMapa['icono_cajero'].append(self.nodo_aux)
                
            elif self.nodo_aux.getTagValue('amenity') == 'bank':
                self.elementosMapa['icono_banco'].append(self.nodo_aux)
                
            elif self.nodo_aux.getTagValue('amenity') == 'pharmacy':
                self.elementosMapa['icono_farmacia'].append(self.nodo_aux)
                
            elif self.nodo_aux.getTagValue('amenity') == 'hospital':
                self.elementosMapa['icono_hospital'].append(self.nodo_aux)
                
            elif self.nodo_aux.getTagValue('amenity') == 'cinema':
                self.elementosMapa['icono_cine'].append(self.nodo_aux)
                
            elif self.nodo_aux.getTagValue('amenity') == 'theatre':
                self.elementosMapa['icono_teatro'].append(self.nodo_aux)
                
            elif self.nodo_aux.getTagValue('amenity') == 'place_of_worship':
                if self.nodo_aux.getTagValue('religion') == 'christian':
                    self.elementosMapa['icono_iglesia'].append(self.nodo_aux)
                if self.nodo_aux.getTagValue('religion') == 'muslim':
                    self.elementosMapa['icono_mezquita'].append(self.nodo_aux)
            
            elif self.nodo_aux.getTagValue('amenity') == 'police':
                self.elementosMapa['icono_policia'].append(self.nodo_aux)    
            
            elif self.nodo_aux.getTagValue('amenity') == 'post_office':
                self.elementosMapa['icono_oficina_correos'].append(self.nodo_aux)
            
            elif self.nodo_aux.getTagValue('tourism') == 'museum':
                self.elementosMapa['icono_museo'].append(self.nodo_aux)
                
            elif self.nodo_aux.getTagValue('tourism') == 'hotel':
                self.elementosMapa['icono_hotel'].append(self.nodo_aux)
            
            
            self.N[ self.nodo_aux.getId() ] = self.nodo_aux
            
            self.nodo_aux = Nodo()
        
        if nombre == 'way':
            
            self.way_aux.addTags(self.tags_aux)
            self.tags_aux = {}
            
            # incluye el way con todos los datos(refs y tags) a la lista
            self.W[self.way_aux.getId()] = self.way_aux
            
            if self.way_aux.getTagValue('highway') == 'motorway':
                # puede ir en combinacion con el resto
                if self.way_aux.getTagValue('junction') == 'roundabout':
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
                else:
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',120) # es un valor implicito
            
                # añade a way a la lista de carreteras
                self.elementosMapa['carreteras'].append(self.way_aux)
                # crea la lista de adyacencia asociado a way
                self.setListaAdyacencia()
            
            elif self.way_aux.getTagValue('highway') == 'motorway_link':
                # puede ir en combinacion con el resto
                if self.way_aux.getTagValue('junction') == 'roundabout':
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
                else:
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
            
                # añade a way a la lista de carreteras
                self.elementosMapa['carreteras'].append(self.way_aux)
                # crea la lista de adyacencia asociado a way
                self.setListaAdyacencia()
            
            elif self.way_aux.getTagValue('highway') == 'trunk':
                # puede ir en combinacion con el resto
                if self.way_aux.getTagValue('junction') == 'roundabout':
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
                else:
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',120) # es un valor implicito
                
                # añade a way a la lista de carreteras
                self.elementosMapa['carreteras'].append(self.way_aux)
                # crea la lista de adyacencia asociado a way
                self.setListaAdyacencia()
            
            elif self.way_aux.getTagValue('highway') == 'trunk_link':
                # puede ir en combinacion con el resto
                if self.way_aux.getTagValue('junction') == 'roundabout':
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
                else:
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
                
                # añade a way a la lista de carreteras
                self.elementosMapa['carreteras'].append(self.way_aux)
                # crea la lista de adyacencia asociado a way
                self.setListaAdyacencia()
            
            elif self.way_aux.getTagValue('highway') == 'primary':
                # puede ir en combinacion con el resto
                if self.way_aux.getTagValue('junction') == 'roundabout':
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
                else:
                    self.way_aux.addTag('maxspeed',100) # es un valor implicito
            
                # añade a way a la lista de carreteras
                self.elementosMapa['carreteras'].append(self.way_aux)
                # crea la lista de adyacencia asociado a way
                self.setListaAdyacencia()
            
            elif self.way_aux.getTagValue('highway') == 'primary_link':
                # puede ir en combinacion con el resto
                if self.way_aux.getTagValue('junction') == 'roundabout':
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
                else:
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
            
                # añade a way a la lista de carreteras
                self.elementosMapa['carreteras'].append(self.way_aux)
                # crea la lista de adyacencia asociado a way
                self.setListaAdyacencia()
            
            elif self.way_aux.getTagValue('highway') == 'secondary':
                # puede ir en combinacion con el resto
                if self.way_aux.getTagValue('junction') == 'roundabout':
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
                else:
                    self.way_aux.addTag('maxspeed',90) # es un valor implicito
                
                # añade a way a la lista de carreteras
                self.elementosMapa['carreteras'].append(self.way_aux)
                # crea la lista de adyacencia asociado a way
                self.setListaAdyacencia()
            
            elif self.way_aux.getTagValue('highway') == 'secondary_link':
                # puede ir en combinacion con el resto
                if self.way_aux.getTagValue('junction') == 'roundabout':
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
                else:
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
                
                # añade a way a la lista de carreteras
                self.elementosMapa['carreteras'].append(self.way_aux)
                # crea la lista de adyacencia asociado a way
                self.setListaAdyacencia()
            
            elif self.way_aux.getTagValue('highway') == 'secondary':
                # puede ir en combinacion con el resto
                if self.way_aux.getTagValue('junction') == 'roundabout':
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
                else:
                    self.way_aux.addTag('maxspeed',90) # es un valor implicito
                
                # añade a way a la lista de carreteras
                self.elementosMapa['carreteras'].append(self.way_aux)
                # crea la lista de adyacencia asociado a way
                self.setListaAdyacencia()
            
            elif self.way_aux.getTagValue('highway') == 'tertiary':
                # puede ir en combinacion con el resto
                if self.way_aux.getTagValue('junction') == 'roundabout':
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
                else:
                    self.way_aux.addTag('maxspeed',90) # es un valor implicito
                
                # añade a way a la lista de carreteras
                self.elementosMapa['carreteras'].append(self.way_aux)
                # crea la lista de adyacencia asociado a way
                self.setListaAdyacencia()
            
            elif self.way_aux.getTagValue('highway') == 'unclassified':
                # puede ir en combinacion con el resto
                if self.way_aux.getTagValue('junction') == 'roundabout':
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
                else:
                    self.way_aux.addTag('maxspeed',90) # es un valor implicito
                
                # añade a way a la lista de carreteras
                self.elementosMapa['carreteras'].append(self.way_aux)
                # crea la lista de adyacencia asociado a way
                self.setListaAdyacencia()
            
            elif self.way_aux.getTagValue('highway') == 'road':
                # puede ir en combinacion con el resto
                if self.way_aux.getTagValue('junction') == 'roundabout':
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
                else:
                    self.way_aux.addTag('maxspeed',90) # es un valor implicito

                # añade a way a la lista de carreteras
                self.elementosMapa['carreteras'].append(self.way_aux)
                # crea la lista de adyacencia asociado a way
                self.setListaAdyacencia()
            
            elif self.way_aux.getTagValue('highway') == 'residential':
                # puede ir en combinacion con el resto
                if self.way_aux.getTagValue('junction') == 'roundabout':
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
                else:
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
            
                # añade a way a la lista de carreteras
                self.elementosMapa['carreteras'].append(self.way_aux)
                # crea la lista de adyacencia asociado a way
                self.setListaAdyacencia()
            
            elif self.way_aux.getTagValue('highway') == 'living_street':
                # puede ir en combinacion con el resto
                if self.way_aux.getTagValue('junction') == 'roundabout':
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
                else:
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
            
                # añade a way a la lista de carreteras
                self.elementosMapa['carreteras'].append(self.way_aux)
                # crea la lista de adyacencia asociado a way
                self.setListaAdyacencia()
            
            elif self.way_aux.getTagValue('highway') == 'service':
                # puede ir en combinacion con el resto
                if self.way_aux.getTagValue('junction') == 'roundabout':
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito
                else:
                    self.way_aux.addTag('maxspeed',25) # es un valor implicito
                
                # puede ir en combinacion con el resto
                if self.way_aux.getTagValue('junction') == 'roundabout':
                    self.way_aux.addTag('oneway','yes') # es un valor implicito
                    self.way_aux.addTag('maxspeed',50) # es un valor implicito

                # añade a way a la lista de carreteras
                self.elementosMapa['carreteras'].append(self.way_aux)
                # crea la lista de adyacencia asociado a way
                self.setListaAdyacencia()
            
            elif self.way_aux.getTagValue('highway') == 'pedestrian':
                if self.way_aux.getTagValue('area') == 'yes':
                    self.elementosMapa['plazas'].append(self.way_aux)
                else:
                    self.elementosMapa['peatonales'].append(self.way_aux)
                 
            elif self.way_aux.getTagValue('junction') == 'roundabout':
                self.way_aux.addTag('oneway','yes') # es un valor implicito
                self.way_aux.addTag('maxspeed',50) # es un valor implicito
                
                # añade a way a la lista de carreteras
                self.elementosMapa['carreteras'].append(self.way_aux)
                # crea la lista de adyacencia asociado a way
                self.setListaAdyacencia()
             
            elif self.way_aux.getTagValue('leisure') == 'garden':
                self.elementosMapa['garden'].append(self.way_aux)
            
            # lista de areas
            
            elif self.way_aux.getTagValue('amenity') == 'restaurant':
                self.elementosMapa['area_restaurante'].append(self.way_aux)
                
            elif self.way_aux.getTagValue('amenity') == 'food_court':
                self.elementosMapa['area_restaurante'].append(self.way_aux)
            
            elif self.way_aux.getTagValue('amenity') == 'fast_food':
                self.elementosMapa['area_restaurante'].append(self.way_aux)
            
            elif self.way_aux.getTagValue('amenity') == 'pub':
                self.elementosMapa['area_bar'].append(self.way_aux)
            
            elif self.way_aux.getTagValue('amenity') == 'bar':
                self.elementosMapa['area_bar'].append(self.way_aux)
                
            elif self.way_aux.getTagValue('amenity') == 'cafe':
                self.elementosMapa['area_cafe'].append(self.way_aux)
                
            elif self.way_aux.getTagValue('amenity') == 'kindergarten':
                self.elementosMapa['area_guarderia'].append(self.way_aux)
            
            elif self.way_aux.getTagValue('amenity') == 'school':
                self.elementosMapa['area_colegio'].append(self.way_aux)
            
            elif self.way_aux.getTagValue('amenity') == 'college':
                self.elementosMapa['area_colegio'].append(self.way_aux)
            
            elif self.way_aux.getTagValue('amenity') == 'library':
                self.elementosMapa['area_libreria'].append(self.way_aux)
            
            elif self.way_aux.getTagValue('amenity') == 'university':
                self.elementosMapa['area_universidad'].append(self.way_aux)
            
            elif self.way_aux.getTagValue('amenity') == 'bus_station':
                self.elementosMapa['area_autobus'].append(self.way_aux)
            
            elif self.way_aux.getTagValue('highway') == 'bus_stop':
                self.elementosMapa['area_autobus'].append(self.way_aux)
            
            elif self.way_aux.getTagValue('amenity') == 'fuel':
                self.elementosMapa['area_gasolinera'].append(self.way_aux)
            
            elif self.way_aux.getTagValue('amenity') == 'parking':
                self.elementosMapa['area_parking'].append(self.way_aux)
                
            elif self.way_aux.getTagValue('amenity') == 'taxi':
                self.elementosMapa['area_taxi'].append(self.way_aux)
            
            elif self.way_aux.getTagValue('amenity') == 'atm':
                self.elementosMapa['area_cajero'].append(self.way_aux)
                
            elif self.way_aux.getTagValue('amenity') == 'bank':
                self.elementosMapa['area_banco'].append(self.way_aux)
                
            elif self.way_aux.getTagValue('amenity') == 'pharmacy':
                self.elementosMapa['area_farmacia'].append(self.way_aux)
                
            elif self.way_aux.getTagValue('amenity') == 'hospital':
                self.elementosMapa['area_hospital'].append(self.way_aux)
                
            elif self.way_aux.getTagValue('amenity') == 'cinema':
                self.elementosMapa['area_cine'].append(self.way_aux)
                
            elif self.way_aux.getTagValue('amenity') == 'theatre':
                self.elementosMapa['area_teatro'].append(self.way_aux)
                
            elif self.way_aux.getTagValue('amenity') == 'place_of_worship':
                if self.way_aux.getTagValue('religion') == 'christian':
                    self.elementosMapa['area_iglesia'].append(self.way_aux)
                if self.way_aux.getTagValue('religion') == 'muslim':
                    self.elementosMapa['area_mezquita'].append(self.way_aux)
            
            elif self.way_aux.getTagValue('amenity') == 'police':
                self.elementosMapa['area_policia'].append(self.way_aux)    
            
            elif self.way_aux.getTagValue('amenity') == 'post_office':
                self.elementosMapa['area_oficina_correos'].append(self.way_aux)
            
            elif self.way_aux.getTagValue('tourism') == 'museum':
                self.elementosMapa['area_museo'].append(self.way_aux)
                
            elif self.way_aux.getTagValue('tourism') == 'hotel':
                self.elementosMapa['area_hotel'].append(self.way_aux)
            
            # edificios
            
            if self.way_aux.getTagValue('building') == 'yes':
                self.elementosMapa['edificios'].append(self.way_aux)
            
            # crea un nuevo way
            self.way_aux = Way()
        
    # crea los enlaces de la matris de camino    
    def setListaAdyacencia(self):
        
        # recorre la lista de nodos estableciendo algunos atributos
        for nd in self.way_aux.getListaRefs():
            nd.CrearListaAdyacencia()
            
            # establece el valor oneway del nodo, 'yes' por defecto
            if self.way_aux.getTagValue('oneway') != 'yes' and self.way_aux.getTagValue('oneway') != '1':
                nd.setOneWay('no')
            
            # establece el nombre de way en el nodo
            wayName = self.way_aux.getTagValue('name')
            if wayName is None:
                wayName = self.way_aux.getTagValue('ref')
                if wayName is None:
                    wayName = '('+self.way_aux.getId()+')'
            nd.addWayName(wayName)    
        
        # recorre la lista de nodos creando la matriz de camino
        for i in range(self.way_aux.getListaRefs().__len__()-1):
            
            # establece el enlace entre el nodo y el siguiente
            self.way_aux.getListaRefs()[i].addListaAdyacencia(self.way_aux.getListaRefs()[i+1])
            
            if self.way_aux.getTagValue('oneway') != 'yes' and self.way_aux.getTagValue('oneway') != '1':
                # establece el enlace entre el nodo y el anterior
                self.way_aux.getListaRefs()[i+1].addListaAdyacencia(self.way_aux.getListaRefs()[i])
        
#class ClienteHTTP(threading.Thread):
#
#    def __init__(self,latitud,longitud):
#
#        threading.Thread.__init__(self)
#
#        conn = httplib.HTTPConnection("www.openstreetmap.org")
#        conn.request("GET",
#                     "/api/0.6/map?bbox=" + str(longitud-0.02) + "," + str(latitud-0.02) + "," + str(longitud+0.02) + "," + str(latitud+0.02))
#        r1 = conn.getresponse()
#        print r1.status, r1.reason # 200, OK
#        data1 = r1.read()
#        print data1
#        http = HTTPClient()
#
#        self.channel = http.getDocument(URLSpec("http://www.openstreetmap.org/api/0.6/map?bbox=" + str(longitud-0.02)
#                                           +"," 
#                                           + str(latitud-0.02)
#                                           +","
#                                           + str(longitud+0.02)
#                                           +","
#                                           +str(latitud+0.02)))
#        
#        
#        self.rf = Ramfile()
#        self.channel.downloadToRam(self.rf)
#        taskMgr.add(self.getHTTP, 'download')
#
#    def getHTTP(self,task):
#
#        if self.channel.run():
#            # Still waiting for file to finish downloading.
#            return task.cont
#        if not self.channel.isDownloadComplete():
#            print "Error downloading file."
#            return task.done
#        self.data1 = self.rf.getData()
#        print self.data1
##        parseString(self.data1,Manejador(self.bounds,self.nodos,self.ways))
#        return task.done



class Mapa():    

    
    def __init__(self,osm="http",escala=Escalas['gr-px'],latitud=36.7,longitud=-6.1):

        print "Mapa"

        # almacena el nombre del mapa
        self.mapa = osm
        # alamcena los limites del sistema
        self.bounds = Bounds()
        # diccionario de nodos por id
        self.nodos = {}  
        # Lista de ways
        self.ways = {}
    
        # diccionario de lista que agrupa los way de cada tipo
        self.elementosMapa = {}
    
        # diccionario de nodos que contienen información
        self.iconosMapa = {}
    
        if self.mapa == "http":
    
            print 'Accediendo a www.openstreetmap.org ...'
            
            # mapa de http client
            conn = httplib.HTTPConnection("www.openstreetmap.org")
            conn.request("GET",
                         "/api/0.6/map?bbox=" + str(longitud-0.02) + "," + str(latitud-0.02) + "," + str(longitud+0.02) + "," + str(latitud+0.02))
            r1 = conn.getresponse()
            print r1.status, r1.reason # 200, OK
            data1 = r1.read()
            parseString(data1,Manejador(self.bounds,self.nodos,self.ways,self.elementosMapa,self.iconosMapa))
              

        else:
            
            # crea un objeto manejador de eventos 
            manejador = Manejador(self.bounds,self.nodos,self.ways,self.elementosMapa,self.iconosMapa)
            # crea un objeto parser
            parser = make_parser()
            # asigna nuestro manejador de evenos al parser
            parser.setContentHandler(manejador)
            # abre el fichero xml con los datos
            fichero = open('./mapas/' + self.mapa, 'r')
            # asigna al parser el fichero con la informacion
            parser.parse(fichero)
        
        print "Mapa:Documento osm procesado"
        
        # crea la martiz de camino
#        self.Mcamino = MatrizCamino(self.nodos,self.carreteras)
#        self.Mcamino = MatrizCamino(self.elementosMapa['carreteras'])



    # devuelve el nombre del mapa
    def getMapa(self):
        return str(self.nombre)
    
    def CalcularPorcionMapa(self,pos,distancia):
        
        print 'Mapa:CalculaPorcionMapa'
    
        """
         Dtermina que elmentos estan en la parte
         visible de la porcion de mapa elegida
        """
    
        # contiene la listas  way clasificados
        self.porcion_mapa = {}
    
        #  items() Return a copy of the dictionary’s list of (key, value) pairs
        for eltos in self.elementosMapa.items():
    
            self.porcion_mapa[eltos[0]] = []
    
            for w in eltos[1]:
    
                for r in w.getListaRefs().__iter__():
                    if (r.getPos() - pos).length() < distancia:
           
                        self.porcion_mapa[eltos[0]].append(w)
           
                        # no es necesario seguir recorriento la lista de referencias
                        break
                        
        """
         Determina la la parte de la matriz de camino
         perteneciente a la parte visible del mapa
        """
        self.porcion_camino = []

        for w in self.elementosMapa['carreteras'].__iter__():
            for nd in w.getListaRefs().__iter__():
                if (nd.getPos() - pos).length() < distancia:
                    self.porcion_camino.append(nd)
                    
    """
    Devuevle una lista de elementos  que estan dentro 
    de la pocion de mapa seleccionada
    """
    def getPorcionMapa(self):
      
        # devuelve la lista de elementos que deben ser pintados
        return self.porcion_mapa
    
    """
    devuelve una lisa de nodos pertenecientes a la matriz de camino
    que estan dentro de la porcion de mapa dibujada
    """
    def getPorcionCamino(self):
        return self.porcion_camino
    
    # devuelve un objeto bounds con los limites del mapa
    def getBounds(self):
        return self.bounds
    def getMinLon(self):
        return self.bounds.getMinLon()
    def getMaxLon(self):
        return self.bounds.getMaxLon()
    def getMinLat(self):
        return self.bounds.getMinLat()
    def getMaxLat(self):
        return self.bounds.getMaxLat()
    # devuelve una tupla con la posicion minima del limite del mapa
    def getBoundsMinX(self):
        return self.bounds.getMinX()
    def getBoundsMinY(self):
        return self.bounds.getMinY()
    # devuelve una tupla con la posicion maxima del limite del mapa
    def getBoundsMaxX(self):
        return self.bounds.getMaxX()
    def getBoundsMaxY(self):
        return self.bounds.getMaxY()    
        
    # devuelve la lista de id de los nodos del mapa
    def getListaNodosId(self):
        return self.nodos.iterkeys()
    # tupla con la posicon del nodo
    def getNodoPosById(self,id):
        return self.nodos[id].getPos()
    # devuelve la coordenada X de la posicion del nodo
    def getNodoXById(self,id):
        return self.nodos[id].getX()
    # devulve la coordenada Y de la posicion del nodo
    def getNodoYById(self,id):
        return self.nodos[id].getY()
    
    def getNodoLonById(self,id):
        return self.nodos[id].getLon() # longitud del nodo    
    def getNodoLatById(self,id):
        return self.nodos[id].getLat() # latitud del nodo
    
    # devuelve la lista de ways
    def getListaWays(self):
        return self.ways.keys()
    
    def getListaWaysByNodo(self,nd):
        return self.nodos[nd].getListaWays()
    
    def getWay(self,id):
        return self.ways[id]
    
    def getWayNombreByNodo(self,n):
        return self.nodos[n].getWayName()
    
    def getWayNombreById(self,w):
        return self.ways[w].getTagValue('name')
    
    def getOneWayByNodo(self,id):
        return self.nodos[id].getOneWay()    
    
#    # metodos de la matriz de camino
#    def getMatrizCamino(self):
#        return self.Mcamino
#    def getListaNodosMatrizCamino(self):
#        return self.Mcamino.getListaNodos()
#    
##    def getListaNodosSiguientes(self,id):
##        return self.Mcamino.getListaNodosSiguientes(id)
##    def getListaNodosAnteriores(self,id):
##        return self.Mcamino.getListaNodosAnteriores(id)
#    
#    def getListaAdyacentes(self,id):
#        return self.Mcamino.getListaAdyacentes(id)

    def getRuta(self,origen,destino):
        
        print 'Mapa:getRuta'
        
        # matriz de costes
        G = {}
        
        # establece los costes iniciales
    
        # recorre la lisa de carreteras
        for way in self.elementosMapa['carreteras']:
            # recorre la lista de nodos de la carretera
            for nodo in way.getListaRefs():
                
                aux = {} 
                # recorre la lista de adyacencia del nodo
                for ad in nodo.getListaAdyacentes():
                    aux[ad.getId()] = (ad.getPos() - nodo.getPos()).length()
                
                G[nodo.getId()] = aux
                                
        try:
            ruta = []
            for id in shortest_path(G,origen.getId(),destino.getId()):
                ruta.append(self.nodos[id])
            return ruta
        except:
            print 'Mapa:getRuta, no fue posible encontrar un camino.'
            return []
        
    def getListaCarreteras(self):
        return self.elementosMapa['carreteras']


#from pandac.PandaModules import HTTPClient
#from pandac.PandaModules import Ramfile
#from pandac.PandaModules import URLSpec
#      
#http = HTTPClient()
#
#latitud=36.7
#longitud=-6.1
#channel = http.getDocument(URLSpec("http://www.openstreetmap.org/api/0.6/map?bbox=" + str(longitud-0.02)
#                                   +"," 
#                                   + str(latitud-0.02)
#                                   +","
#                                   + str(longitud+0.02)
#                                   +","
#                                   +str(latitud+0.02)))
#
#while channel.run():
#    print channel.getStatusString()
#
##if not channel.isDownloadComplete():
##    print 'error'
#
##channel.setConnectTimeout(15)
##channel.setHttpTimeout(15)
#rf = Ramfile()
#channel.downloadToRam(rf)
#
#if not channel.isValid():
#    print 'error2' 
# 
#data = rf.getData()
#print "got data:"
##print data

        
#M = Mapa('cadiz.osm')
#
#nodo_destino = '84807716'
#print nodo_destino
#print 'MatrizCamino1',M.getListaNodosSiguientes(nodo_destino),M.getListaNodosAnteriores(nodo_destino)
#print 'MatrizCamino',M.getListaAdyacentes(nodo_destino)
            
#    
#M = Mapa("cadiz.osm")
#
#print 'inicio'
#print M.Ruta("298900678","324013688")
#print 'fin'

#print 'total ways',M.getListaWays().__len__()
#print 'inicio'
#print M.getPorcionMapa(VBase3(0,0,0),Escalas['m-px']*500).keys().__len__()
#print 'fin'

#camino = M.getMatrizCamino()
#for p in camino.Mcamino:
#    if p == '259181304':
#        print 'nodo',p
#        print 'Siguientes: ',M.getListaNodosSiguientes(p)
#        print 'Anteriores: ',M.getListaNodosAnteriores(p)
#        print '------------------------------------------'
#    if p == '309847062':
#        print 'nodo',p
#        print 'Siguientes: ',M.getListaNodosSiguientes(p)
#        print 'Anteriores: ',M.getListaNodosAnteriores(p)
#        print '------------------------------------------'
#""" 
#
#"""
#lista = [1,2,3,4,5,6,7,8]
## todos menos el ultimo
#print lista[:-1]
## print todos menos el primero
#print lista[1:]
#"""


