# -*- coding: utf-8 -*-

# DirectStart contiene los modulos de Panda3d
#import direct.directbase.DirectStart

from pandac.PandaModules import NodePath

# para configurar propiedades de la ventana
from pandac.PandaModules import WindowProperties

# metodo rapido para texno en pantalla
from direct.gui.OnscreenText import OnscreenText

# para cargar imagenes
from direct.gui.OnscreenImage import OnscreenImage

# para la entrada de texto
from direct.gui.DirectEntry import DirectEntry

# para crear un objeto texto y poder incluirlo luego en un NodePath
from pandac.PandaModules import TextNode

# para capturar los eventos de teclado
from direct.showbase.DirectObject import DirectObject

#Load models, textures, sounds, and code.
from direct.showbase.Loader import Loader

# para el bucle de codigo del videojuego
#from direct.task import Task
#from pandac.PandaModules import Vec3

#para triangular y dibujar areas cerradas
from pandac.PandaModules import Triangulator
from pandac.PandaModules import GeomVertexArrayFormat
from pandac.PandaModules import InternalName
from pandac.PandaModules import Geom
from pandac.PandaModules import GeomVertexFormat
from pandac.PandaModules import GeomVertexData
from pandac.PandaModules import GeomVertexWriter
from pandac.PandaModules import GeomTriangles
from pandac.PandaModules import GeomNode
from pandac.PandaModules import GeomLines
from pandac.PandaModules import GeomTristrips

from pandac.PandaModules import Texture
from pandac.PandaModules import TextureStage
from pandac.PandaModules import TexGenAttrib
from pandac.PandaModules import TransparencyAttrib 

# para crear camaras
from pandac.PandaModules import Camera
# para confugirar la lente de la camara
from pandac.PandaModules import OrthographicLens

from direct.task.Task import Task

from pandac.PandaModules import TextNode
from pandac.PandaModules import TextFont

from pandac.PandaModules import CollisionTraverser
from pandac.PandaModules import CollisionHandlerEvent
from pandac.PandaModules import CollisionNode
from pandac.PandaModules import CollisionInvSphere
from pandac.PandaModules import CollisionPlane
from pandac.PandaModules import CollisionSphere
from pandac.PandaModules import BitMask32

from pandac.PandaModules import Vec3
from pandac.PandaModules import VBase3
from pandac.PandaModules import VBase2
from pandac.PandaModules import Point2D
from pandac.PandaModules import Point3
from pandac.PandaModules import Plane

from math import atan 
from math import degrees
from math import sqrt
from math import ceil
from math import radians
from math import cos
from math import sin

from random import randint

from re import match,IGNORECASE

try:
    import cPickle as pickle
except ImportError:
    import pickle   

from mapa import Mapa
from circulacion import Circulacion
from escalas import Escalas


# clase generica para todos los elementos del escenario
class Figura:
    
    def Puente(self,puente,ax,ay,bx,by,oneway='no'):

        # calculo del punto medio
        puenteX = (ax + bx)/2
        puenteY = (ay + by)/2
        puente.setPos(puenteX,puenteY,0)      
            
        #calculo del angulo de giro
        if(ax == bx):
            puente.setH(90)
            
        elif(ay == by):
            puente.setH(0)
        else:
            angulo = abs(ay-by)/abs(ax-bx)
            anguloRad = atan(angulo) 
            anguloDegrees = degrees(anguloRad)
        
            vx = bx-ax
            vy = by-ay
            if vx<0 and vy>0: # segundo cuadrante
                anguloDegrees = 180 - anguloDegrees
            elif vx<0 and vy<0: # tercer cuadrante
                anguloDegrees = anguloDegrees + 180
            elif vx>0 and vy<0: # cuarto cuadrante
                anguloDegrees = anguloDegrees * (-1)
        
            puente.setHpr(anguloDegrees,0,0)
        
        
        #calculo de la escala, dividido por 100 xq la pieza mide 100 pixeles
        escala = sqrt( pow( ax-bx ,2) + pow( ay-by ,2) ) /32
        if oneway != 'yes':
            puente.setScale(escala,Escalas['m-px'] * 6.0 /32.0,1)
        else:
            puente.setScale(escala,Escalas['m-px'] * 3.0 /32.0,1)
        
    def setTextura(self, np, seleccion, uScale=0.5, vScale=0.5, WrapU=Texture.WMMirror, \
                   WrapV=Texture.WMMirror, TexGen=TexGenAttrib.MWorldPosition):
        
        # carga una textura
        textura = loader.loadTexture('texturas/'+ seleccion+'.png')
        textura.setWrapU(WrapU)
        textura.setWrapV(WrapV)
        #textura.setBorderColor(VBase4(1, 1, 1, 0.0))
        # crea las propiedades de la textura
        stage = TextureStage('ts')
        # estable la coordianacion de los vertices u,v
        stage.setTexcoordName("coord")
            
        # establce la textura(propiedades, textura, 1 idica prioridad sobre la textura anterior)
        np.setTexture(stage, textura, 1)
        # realiza la coordinacion de vertices
        np.setTexGen(stage, TexGen)
        #establece la escala de la textura
        np.setTexScale(stage,uScale,vScale)
        # desplaza la textura sobre el modelo
        #np.setTexPos(stage,-0.39,-0.9,0)
        # rota la textura
        #np.setTexRotate(stage,-15)
        
    def setNombreIcono(self, nombre='Nombre', pos=VBase3(0, 0, 0),\
                       altura=0, parent=render):
    
        fontSansBold = loader.loadFont('fuentes/FreeSansBold.ttf')
        prioridad = 150
        pos = VBase3(pos)
        pos.setZ(altura)
        
        texto = TextNode('texto')
        texto.setText(nombre)
        texto.setFont(fontSansBold)
        texto.setAlign( TextNode.ACenter )
        texto.setTextColor(0,0,0,0.6)
        texto.setCardColor(1, 1, 1, 0.5)
        texto.setCardAsMargin(0, 0, 0, 0)
        texto.setCardDecal(True)
        
        # letrero 1
        TNodePath = NodePath(texto)
        TNodePath.setScale(10)
        TNodePath.setH(0)
        TNodePath.setPos(pos)
        TNodePath.setBin("fixed", prioridad)
        TNodePath.reparentTo(parent)
        
        self.modelos.append(TNodePath)

        # letrero 2
        texto = TextNode('texto')
        texto.setText(nombre)
        texto.setFont(fontSansBold)
        texto.setAlign( TextNode.ACenter )
        texto.setTextColor(0,0,0,0.6)
        texto.setCardColor(1, 1, 1, 0.5)
        texto.setCardAsMargin(0, 0, 0, 0)
        texto.setCardDecal(True)

        TNodePath = NodePath(texto)
        TNodePath.setScale(10)
        TNodePath.setH(90)
        TNodePath.setPos(pos)
        TNodePath.setBin("fixed", prioridad)
        TNodePath.reparentTo(parent)
        
        self.modelos.append(TNodePath)

        # letrero 3
        texto = TextNode('texto')
        texto.setText(nombre)
        texto.setFont(fontSansBold)
        texto.setAlign( TextNode.ACenter )
        texto.setTextColor(0, 0, 0, 0.6)
        texto.setCardColor(1, 1, 1, 0.5)
        texto.setCardAsMargin(0, 0, 0, 0)
        texto.setCardDecal(True)

        TNodePath = NodePath(texto)
        TNodePath.setScale(10)
        TNodePath.setH(180)
        TNodePath.setPos(pos)
        TNodePath.setBin("fixed", prioridad)
        TNodePath.reparentTo(parent)
        
        self.modelos.append(TNodePath)

        # letrero 4
        texto = TextNode('texto')
        texto.setText(nombre)
        texto.setFont(fontSansBold)
        texto.setAlign( TextNode.ACenter )
        texto.setTextColor(0, 0, 0, 0.6)
        texto.setCardColor(1, 1, 1, 0.5)
        texto.setCardAsMargin(0, 0, 0, 0)
        texto.setCardDecal(True)
        
        TNodePath = NodePath(texto)
        TNodePath.setScale(10)
        TNodePath.setH(270)
        TNodePath.setPos(pos)
        TNodePath.setBin("fixed", prioridad)
        TNodePath.reparentTo(parent)
        
        self.modelos.append(TNodePath)
    
    def setLineaTierra(self, pos=Point3(0, 0, 0), altura=0, parent=render):
        
        #formato manual
        #crea el array de formato
        array = GeomVertexArrayFormat()
        array.addColumn(InternalName.make('vertex'), 3, 
                        Geom.NTFloat32, Geom.CPoint)
        
        #crea el formato e incluye el array de formato
        format = GeomVertexFormat()
        format.addArray(array)
        
        # registra el formato para indicar que se ha establecido
        format = GeomVertexFormat.registerFormat(format)
        
        #crea el objeto Geomvertexdata con el formato especifiado,
        # parametros: nombre identificatico,formato y hint que indica que el objeto es estatico
        vdata = GeomVertexData("vertices", format, Geom.UHStatic)
        
        # se debe crear un GeomVertexWriter por cada columna de GeomVertexData,
        # para rellenar los datos del vertice, algunos nombres de columnas estan reservados
        # vertex para la posicion de los vertices
        vertexWriter = GeomVertexWriter(vdata, 'vertex')
                      
        #anñade los vertices del poligono  
        vertexWriter.addData3f(pos)
        vertexWriter.addData3f(pos + Point3(0, 0, altura))
        
        # define la primitiva apra crear segmentos
        primitiva = GeomLines(Geom.UHStatic)
        
        # añade dos vertices por si numeracion
        primitiva.addVertex(0)
        primitiva.addVertex(1)
        # enlace los vertices añadidos definiendo el segmento
        primitiva.closePrimitive
        
        #crea un objeto Geom e incluye la primitiva anterior que contiene los triangulos
        squareGeom=Geom(vdata)
        squareGeom.addPrimitive(primitiva)
        
        # crea el objeto GeonNode para incluirlo al arbol de rederizar
        # define el nombre "square" para el nodo
        squareGN=GeomNode("square")
        # incluye el objeto geom en el nodo de renderizado GeomNode
        squareGN.addGeom(squareGeom)
        # coloca el objeto GeomNode en el arbol de renderizado
        node_path = NodePath(squareGN) 
        node_path.setRenderModeThickness(1) 
    
        node_path.setBin('fixed',150)
        node_path.reparentTo(parent)
        self.modelos.append(node_path)

    def setAvatar(self, nombre, imagen, pos, altura, parent):
        
        # carga el objeto cubo
        avatar = loader.loadModel('iconos/cubo64')
        avatar.setPos(pos + VBase3(0, 0, altura))
        avatar.setScale(Escalas['m-px'] * 7 / 64)
        avatar.setTransparency(TransparencyAttrib.MAlpha)
        avatar.setColor(1, 1, 1, 0.6)
        avatar.reparentTo(parent)
        avatar.setBin('fixed',150)
        
        # establece la imagen del cubo
        self.setImagenAvatar(avatar, imagen)
        
        self.modelos.append(avatar)

        # crea el letrero con el nombre del icono
        if nombre is not None:
            self.setNombreIcono(nombre, pos, altura*2/3, parent)
        
        # crea la linea a tierra
        self.setLineaTierra(pos, altura, parent)
        
    def setImagenAvatar(self,avatar,imagen):
        
        textura = loader.loadTexture('iconos/'+imagen+'.png')
        avatar.setTexture(textura, 1) 
        
    def getCentroide(self,way): 
        
        c = VBase3(0, 0, 0)
        for nd in way.getListaRefs()[:-1]:
            c += nd.getPos()
        c = c/(way.getListaRefs().__len__()-1)
        
        return c
    
    #recibe una lista de vertices con posiciones (x,y) en el escenario en pixeles, 
    #y el tamano de la lista, devuelve un NodePath con el poligono creado
    def Triangular(self,ListaRefs,parent): 
         
        trig = Triangulator()

        #formato manual
        #crea el array de formato
        array = GeomVertexArrayFormat()
        array.addColumn(InternalName.make('vertex'), 3, 
                        Geom.NTFloat32, Geom.CPoint)
        array.addColumn(InternalName.make('texcoord'), 2, 
                        Geom.NTFloat32, Geom.CTexcoord)
        
        #crea el formato e incluye el array de formato
        format = GeomVertexFormat()
        format.addArray(array)
        
        # registra el formato para indicar que se ha establecido
        format = GeomVertexFormat.registerFormat(format)
        
        #crea el objeto Geomvertexdata con el formato especifiado,
        # parametros: nombre identificatico,formato y hint que indica que el objeto es estatico
        vdata = GeomVertexData("vertices", format, Geom.UHStatic)
        
        # se debe crear un GeomVertexWriter por cada columna de GeomVertexData,
        # para rellenar los datos del vertice, algunos nombres de columnas estan reservados
        # vertex para la posicion de los vertices
        vertexWriter = GeomVertexWriter(vdata, 'vertex')
        
        # segunda columna, para la textura
        texcoordWriter = GeomVertexWriter(vdata, 'texcoord')
    
        
        # añade los datos de la columna vertex, inserta cuatro filas -> cuatro nodos
        
        for ref in ListaRefs[:-1]:
            
            #crea el vertice del poligono  
            vertexWriter.addData3f(ref.getPos())
            #registra el vertice para la triangulacion
            vi = trig.addVertex(ref.getX(), ref.getY())
            # estable el vertice anterior como perteneciente al poligono
            trig.addPolygonVertex(vi)    
            #datos de la textura
            texcoordWriter.addData2f(0,0)
 
        # cada punto significa que habra n unidades de la imagen repetida
        
        #proceso de triangulacion
        trig.triangulate()
        
        #crea la primitiva, en este caso GeomTriangles
        # el paramtro GeomUHStatic define la primitiva com estatica
        
        # para un triangulo GeomTriangles se deben especificar tres vertices y luego ejecutar esta instruccion
        # otro tipo de triangulo 
        primitiva = GeomTriangles(Geom.UHStatic)
        
        #have to add vertices one by one since they are not in order
        for i in range(trig.getNumTriangles()):
            #permite insertar varios vertices a la vez
            primitiva.addVertices(trig.getTriangleV0(i),
                             trig.getTriangleV1(i),
                             trig.getTriangleV2(i))
            primitiva.closePrimitive() 
        
        #crea un objeto Geom e incluye la primitiva anterior que contiene los triangulos
        squareGeom=Geom(vdata)
        squareGeom.addPrimitive(primitiva)
        
        # crea el objeto GeonNode para incluirlo al arbol de rederizar
        # define el nombre "square" para el nodo
        squareGN=GeomNode("square")
        # incluye el objeto geom en el nodo de renderizado GeomNode
        squareGN.addGeom(squareGeom)
        # coloca el objeto GeomNode en el arbol de renderizado
        node_path = parent.attachNewNode(squareGN) 
        
        return node_path
            
    def reparentTo(self,parent):
        for i in self.modelos:
            i.reparentTo(parent)
        
    def __del__(self):
        
        for i in self.modelos:
            i.removeNode()    

class Icono(Figura):
    
    def __init__(self,nodo,parent,imagen):
        
        self.modelos = []
        
        altura = Escalas['m-px'] * 7
        
        # crea el cartel con el nombre del icono
        self.setAvatar(nodo.getTagValue('name'), imagen, nodo.getPos(), \
                       altura, parent)
        
#        # imagen del suelo
#        marca = OnscreenImage(image='imagenes/icono_nombre.png',
#                              pos=nodo.getPos(),
#                              scale=Escalas['m-px'] * 3,
#                              parent=parent)
#        
#        marca.setTransparency(TransparencyAttrib.MAlpha)
#        marca.setBin("fixed", 130)
#        marca.setHpr(0, -90, 0)
#        self.modelos.append(marca)
     
class Edificio(Figura):
    
    def __init__(self,way,parent):
        
        self.modelos = []
        
        orientacion = NodePath('orientacion')
        orientacion.reparentTo(parent)
        
        altura = Escalas['m-px']*4
        
        for i in range(len(way.getListaRefs())-1):
            x = (way.getListaRefs()[i].getPos() - way.getListaRefs()[i+1].getPos()).length()
            y = altura

            np = self.getPared((VBase3(0,0,0), VBase3(x,0,0), 
                                VBase3(x,y,0), VBase3(0,y,0)), 
                                parent)
            
            np.setPos(way.getListaRefs()[i].getPos())
            orientacion.setPos(way.getListaRefs()[i+1].getPos())
            np.lookAt(orientacion)
            np.setP(90)
            np.setH(np.getH()+90)
            np.setBin('fixed',150)
            np.setTwoSided(True)
            np.reparentTo(parent)
            
            self.setTextura(np, 'textura2', uScale=0.01, vScale=0.01)
            
            self.modelos.append(np)
        
        
        techo = self.getTecho(way.getListaRefs(), altura, parent)
        techo.setBin('fixed',150)
        self.setTextura(techo, 'textura2', uScale=0.01, vScale=0.01)
        self.modelos.append(techo)
        
        if way.getTagValue('name') is not None:
            self.setNombreIcono(way.getTagValue('name'), self.getCentroide(way), \
                                Escalas['m-px'] * 8, parent)
           
            self.setLineaTierra(self.getCentroide(way), Escalas['m-px'] * 8, parent) 
        
    #recibe una lista de vertices con posiciones (x,y) en el escenario en pixeles, 
    #y el tamano de la lista, devuelve un NodePath con el poligono creado
    def getPared(self,ListaRefs,parent): 
         
        trig = Triangulator()

        #formato manual
        array = GeomVertexArrayFormat()
        # columna de vertices
        array.addColumn(InternalName.make('vertex'), 3, 
                        Geom.NTFloat32, Geom.CPoint)
        # columna de vertices uv
        array.addColumn(InternalName.make('texcoord'), 2, 
                        Geom.NTFloat32, Geom.CTexcoord)
        
        #crea el formato e incluye el array de formato
        format = GeomVertexFormat()
        format.addArray(array)
        
        # registra el formato para indicar que se ha establecido
        format = GeomVertexFormat.registerFormat(format)
        
        #crea el objeto Geomvertexdata con el formato especifiado,
        # parametros: nombre identificatico,formato y hint que indica que el objeto es estatico
        vdata = GeomVertexData("vertices", format, Geom.UHStatic)
        
        # se debe crear un GeomVertexWriter por cada columna de GeomVertexData,
        # para rellenar los datos del vertice, algunos nombres de columnas estan reservados
        # vertex para la posicion de los vertices
        vertexWriter = GeomVertexWriter(vdata, 'vertex')
        
        #segunda columna, para la textura
        texcoordWriter = GeomVertexWriter(vdata, 'texcoord')
        
        # añade los datos de la columna vertex, inserta cuatro filas -> cuatro nodos
        
        for ref in ListaRefs:
            #crea el vertice del poligono  
            vertexWriter.addData3f(ref)
            #registra el vertice para la triangulacion
            vi = trig.addVertex(ref.getX(),ref.getY())
            # estable el vertice anterior como perteneciente al poligono
            trig.addPolygonVertex(vi)    
            #datos de la textura
            texcoordWriter.addData2f(0,0)
 
        # cada punto significa que habra n unidades de la imagen repetida
        
        #proceso de triangulacion
        trig.triangulate()
        
        #crea la primitiva, en este caso GeomTriangles
        # el paramtro GeomUHStatic define la primitiva com estatica
        # para un triangulo GeomTriangles se deben especificar tres vertices y luego ejecutar esta instruccion
        # otro tipo de triangulo 
        primitiva = GeomTriangles(Geom.UHStatic)
        
        #have to add vertices one by one since they are not in order
        for i in range(trig.getNumTriangles()):
            #permite insertar varios vertices a la vez
            primitiva.addVertex(trig.getTriangleV0(i))
            primitiva.addVertex(trig.getTriangleV1(i))
            primitiva.addVertex(trig.getTriangleV2(i))
            primitiva.closePrimitive()
        primitiva.closePrimitive()
        
        #crea un objeto Geom e incluye la primitiva anterior que contiene los triangulos
        squareGeom=Geom(vdata)
        squareGeom.addPrimitive(primitiva)
        
        # crea el objeto GeonNode para incluirlo al arbol de rederizar
        # define el nombre "square" para el nodo
        squareGN=GeomNode("square")
        # incluye el objeto geom en el nodo de renderizado GeomNode
        squareGN.addGeom(squareGeom)
        # coloca el objeto GeomNode en el arbol de renderizado
        node_path = parent.attachNewNode(squareGN) 
        
        return node_path
    
    def getTecho(self,ListaRefs, altura, parent): 

        techo = self.Triangular(ListaRefs, parent)
        techo.setZ(altura)
        techo.reparentTo(parent)
        
        return techo
    
#    def setTextura(self,seleccion):
#        
#        # carga una textura
#        textura = loader.loadTexture('texturas/'+ seleccion)
#        textura.setWrapU(Texture.WMMirror)
#        textura.setWrapV(Texture.WMMirror)
#        #textura.setBorderColor(VBase4(1, 1, 1, 0.0))
#        
#        for modelo in self.modelos:
#            
#            # crea las propiedades de la textura
#            stage = TextureStage('ts')
#            # estable la coordianacion de los vertices u,v
#            stage.setTexcoordName("coord")
#            # establce la textura(propiedades, textura, 1 idica prioridad sobre la textura anterior)
#            modelo.setTexture(stage, textura, 1)
#            # realiza la coordinacion de vertices
#            modelo.setTexGen(stage, TexGenAttrib.MWorldPosition)
#            #establece la escala de la textura
#            modelo.setTexScale(stage,0.001,0.001)
#            # desplaza la textura sobre el modelo
#            #modelo.setTexPos(stage,-0.39,-0.9,0)
#            # rota la textura
##            modelo.setTexRotate(stage,90)

class Highway(Figura):
    
    class Id(Figura):
        
        def __init__(self,way,parent):
            
            self.modelos = []
            
            for ref in way.getListaRefs():
                
                print 'ref' + ref.getId()
                
                texto = TextNode('texto')
                texto.setText(ref.getId())
                texto.setAlign( TextNode.ACenter )
                texto.setTextColor(0,0,0,1)
                texto.setCardColor(1, 1, 1, 1)
                texto.setCardAsMargin(0, 0, 0, 0)
                texto.setCardDecal(True)
                
                # letrero 1
                TNodePath = NodePath(texto)
                
                # posicion
                TNodePath.setPos(ref.getPos())
#                TNodePath.setZ(Escalas['m-px']*3)
                TNodePath.setScale(10)
                TNodePath.setP(-90)
                
                TNodePath.setBin("fixed", 150)
                TNodePath.reparentTo(parent)
                
                self.modelos.append(TNodePath)
            
    
    def __init__(self,way,mapa,parent):
        
        # lista de nodepath utilizados
        self.modelos = []
        # nodo utilizado para orientar las carreteras
        self.NPOrientacion = NodePath('orientacion')
        
        prioridad = 50
        
        if way.getTagValue('oneway') == 'yes':
            self.UnCarril(way,parent,prioridad)
        else:
            self.DosCarriles(way,mapa,parent,prioridad)

        self.NPOrientacion.removeNode()

    def UnCarril(self,way,parent,prioridad):
        
        # ref: todos menos el ultimo
        # ref+1: todos menos el primero        
        for ref in range(way.getListaRefs()[:-1].__len__()):
            
            nodo1_oneway = way.getListaRefs()[ref].getOneWay()
            nodo2_oneway = way.getListaRefs()[ref+1].getOneWay()
            
            pos1 = way.getListaRefs()[ref].getPos()
            pos2 = way.getListaRefs()[ref+1].getPos()
            
            if nodo1_oneway == 'no' or nodo2_oneway == 'no':
                tramo = self.CalcularPosicionesCarril(pos1 ,pos2)
                
                if nodo1_oneway == 'no':
                    pos1 = tramo[0]
                    
                if nodo2_oneway == 'no':
                    pos2 = tramo[1]
            
            modelo = loader.loadModel('modelos/carretera32')
            acera = loader.loadModel('modelos/acera32')
            
            # se situa en el punto medio de as posiciones
            modelo.setPos((pos1+pos2)/2)
            acera.setPos((pos1+pos2)/2)

            # orienta la carrtera
            self.NPOrientacion.setPos(pos2)
            modelo.lookAt(self.NPOrientacion)
            acera.lookAt(self.NPOrientacion)

            # escala (largo, ancho, alto)
            escala = (pos1-pos2).length()/32
            modelo.setScale(Escalas['m-px']*3/32, escala,1)
            acera.setScale(Escalas['m-px']*3.5/32, escala,1)
 
            modelo.setDepthTest(False)
            modelo.setBin("fixed", prioridad)
            modelo.reparentTo(parent)
            self.modelos.append(modelo)

            acera.setDepthTest(False)
            acera.setBin("fixed", prioridad-1)
            acera.reparentTo(parent)
            self.modelos.append(acera)

            # NODO DE UNION
            union = loader.loadModel("modelos/circulo32") 
            acera = loader.loadModel("modelos/acera_circulo32")
            
            # posicion
            union.setPos( pos1 )
            acera.setPos( pos1 )
            
            # escala
            union.setScale(Escalas['m-px'] * 3 /32)
            acera.setScale(Escalas['m-px'] * 3.5 /32)
            
            union.setDepthTest(False)
            union.setBin("fixed", prioridad+5)
            union.reparentTo(parent)
            self.modelos.append(union)

            acera.setDepthTest(False)
            acera.setBin("fixed", prioridad-1)
            acera.reparentTo(parent)
            self.modelos.append(acera)
            
        # end-for
        
        # ULTIMO NODO DE UNION
        union = loader.loadModel("modelos/circulo32") 
        acera = loader.loadModel("modelos/acera_circulo32")
        
        # posicion
        union.setPos( pos2 )
        acera.setPos( pos2 )
        
        # escala
        union.setScale(Escalas['m-px'] * 3 /32)
        acera.setScale(Escalas['m-px'] * 3.5 /32)
        
        union.setDepthTest(False)
        union.setBin("fixed", prioridad+5)
        union.reparentTo(parent)
        self.modelos.append(union)

        acera.setDepthTest(False)
        acera.setBin("fixed", prioridad-1)
        acera.reparentTo(parent)
        self.modelos.append(acera)

    def DosCarriles(self,way,mapa,parent,prioridad):
        
        # ref: todos menos el ultimo
        # ref+1: todos menos el primero  
        
        primero = True      
        for ref in range(way.getListaRefs()[:-1].__len__()):
            
            nodo1_oneway = way.getListaRefs()[ref].getOneWay()
            nodo2_oneway = way.getListaRefs()[ref+1].getOneWay()
            
            pos1 = way.getListaRefs()[ref].getPos()
            pos2 = way.getListaRefs()[ref+1].getPos()
            
            modelo = loader.loadModel('modelos/carretera32')
            acera = loader.loadModel('modelos/acera32')
            lineas = loader.loadModel('modelos/lineas2carril32')
            
            # se situa en el punto medio de as posiciones
            modelo.setPos((pos1+pos2)/2)
            acera.setPos((pos1+pos2)/2)
            lineas.setPos((pos1+pos2)/2)

            # orienta la carrtera
            self.NPOrientacion.setPos(pos2)
            modelo.lookAt(self.NPOrientacion)
            acera.lookAt(self.NPOrientacion)
            lineas.lookAt(self.NPOrientacion)

            # escala (largo, ancho, alto)
            escala = (pos1-pos2).length()/32
            modelo.setScale(Escalas['m-px']*6/32, escala,1)
            acera.setScale(Escalas['m-px']*6.5/32, escala,1)
            lineas.setScale(Escalas['m-px']*6/32, escala,1)
 
            modelo.setDepthTest(False)
            modelo.setBin("fixed", prioridad)
            modelo.reparentTo(parent)
            self.modelos.append(modelo)

            acera.setDepthTest(False)
            acera.setBin("fixed", prioridad-1)
            acera.reparentTo(parent)
            self.modelos.append(acera)

            lineas.setDepthTest(False)
            lineas.setBin("fixed", prioridad+1)
            lineas.setTransparency(TransparencyAttrib.MAlpha)
            lineas.reparentTo(parent)
            self.modelos.append(lineas)

            # NODO DE UNION
            union = loader.loadModel("modelos/circulo32") 
            acera = loader.loadModel("modelos/acera_circulo32")
            
            # posicion
            union.setPos( pos1 )
            acera.setPos( pos1 )
            
            # escala
            union.setScale(Escalas['m-px'] * 6 /32)
            acera.setScale(Escalas['m-px'] * 6.5 /32)
            
            union.setDepthTest(False)
            
            # si el nodo es un punto de union
#            if mapa.getListaAdyacentes(way.getListaRefs()[ref].getId()).__len__() > 2:
            if way.getListaRefs()[ref].getListaAdyacentes().__len__() > 2:  
                union.setBin("fixed", prioridad+2)
            else:
                union.setBin("fixed", prioridad)
                
            union.reparentTo(parent)
            self.modelos.append(union)

            acera.setDepthTest(False)
            acera.setBin("fixed", prioridad-1)
            acera.reparentTo(parent)
            self.modelos.append(acera)
            
        # end-for
        
        # ULTIMO NODO DE UNION
        union = loader.loadModel("modelos/circulo32") 
        acera = loader.loadModel("modelos/acera_circulo32")
        
        # posicion
        union.setPos( pos2 )
        acera.setPos( pos2 )
        
        # escala
        union.setScale(Escalas['m-px'] * 6 /32)
        acera.setScale(Escalas['m-px'] * 6.5 /32)
        
        union.setDepthTest(False)
        # si el nodo es un punto de union
#        if mapa.getListaAdyacentes(way.getListaRefs()[ref].getId()).__len__() > 2:
        if way.getListaRefs()[ref].getListaAdyacentes().__len__() > 2: 
            union.setBin("fixed", prioridad+2)
        else:
            union.setBin("fixed", prioridad)
        union.reparentTo(parent)
        self.modelos.append(union)

        acera.setDepthTest(False)
        acera.setBin("fixed", prioridad-1)
        acera.reparentTo(parent)
        self.modelos.append(acera)


    # calcula la posicion de partida de cada carril
    # formado para cada par de nodos
    def CalcularPosicionesCarril(self,pos1,pos2):
        
        # angulo desde pos1 a pos 2
        vxy = pos2 - pos1
        
        if(vxy.getX() == 0):
            anguloDegrees = 90
            #self.setHeading(90)
        elif(vxy.getY() == 0):
            anguloDegrees = 0
            #self.setHeading(0)
        else:
            angulo = abs( vxy.getY() / vxy.getX() ) 
            anguloRad = atan(angulo)
            anguloDegrees = degrees(anguloRad)
        # el calculo de los angulos se puede trasladar a cuando
        # a cuando se direge a su nuevo destino
        if vxy.getX()<=0 and vxy.getY()>0: # segundo cuadrante
            anguloDegrees = 180 - anguloDegrees
        elif vxy.getX()<=0 and vxy.getY()<=0: # tercer cuadrante
            anguloDegrees = anguloDegrees + 180
        elif vxy.getX()>0 and vxy.getY()<=0: # cuarto cuadrante
            anguloDegrees = anguloDegrees * (-1)
        
        # desplazamiento desde pos1
        alpha = anguloDegrees - 90
        betaRadians = radians(alpha)
        
        desplaz_carril = Escalas['m-px']*1.2
        
        pos1 =  ( Point3(pos1[0]+cos(betaRadians)*desplaz_carril , pos1[1]+sin(betaRadians)*desplaz_carril,0) )
        
        #desplazamiento desde pos2
        #alpha = anguloDegrees + 90
        #omegaRadians = radians(alpha)
        #pos2 =  ( (pos2[0]+cos(omegaRadians)*3.2 , pos2[1]+sin(omegaRadians)*3.2) )
        
        pos2 =  ( Point3(pos2[0]+cos(betaRadians)*desplaz_carril , pos2[1]+sin(betaRadians)*desplaz_carril,0) )
        
        return (pos1,pos2)

class Peatonal(Figura):
    
    def __init__(self,way,parent):
        
        # lista de nodepath utilizados
        self.modelos = []
        # nodo utilizado para orientar las carreteras
        self.NPOrientacion = NodePath('orientacion')
        
        prioridad = 40
        
        self.Crear(way, parent, prioridad)

        self.NPOrientacion.removeNode()
        
        if way.getTagValue('name') is not None:
            self.setNombreIcono(way.getTagValue('name'), way.getListaRefs()[0].getPos(), \
                                Escalas['m-px'] * 8, parent)
           
            self.setLineaTierra(way.getListaRefs()[0].getPos(), Escalas['m-px'] * 8, parent)

        if way.getTagValue('name') is not None:
            self.setNombreIcono(way.getTagValue('name'), way.getListaRefs()[-1].getPos(), \
                                Escalas['m-px'] * 8, parent)
           
            self.setLineaTierra(way.getListaRefs()[-1].getPos(), Escalas['m-px'] * 8, parent)

    def Crear(self,way,parent,prioridad):
        
        # ref: todos menos el ultimo
        # ref+1: todos menos el primero        
        for ref in range(way.getListaRefs()[:-1].__len__()):
            
            nodo1_oneway = way.getListaRefs()[ref].getOneWay()
            nodo2_oneway = way.getListaRefs()[ref+1].getOneWay()
            
            pos1 = way.getListaRefs()[ref].getPos()
            pos2 = way.getListaRefs()[ref+1].getPos()
            
            modelo = loader.loadModel('modelos/carretera32')
            
            # se situa en el punto medio de as posiciones
            modelo.setPos((pos1+pos2)/2)
            
            # orienta la carrtera
            self.NPOrientacion.setPos(pos2)
            modelo.lookAt(self.NPOrientacion)
            
            # escala (largo, ancho, alto)
            escala = (pos1-pos2).length()/32
            modelo.setScale(Escalas['m-px']*3/32, escala,1)
            
            modelo.setDepthTest(False)
            modelo.setBin("fixed", prioridad)
            modelo.reparentTo(parent)
            self.modelos.append(modelo)

            # NODO DE UNION
            union = loader.loadModel("modelos/circulo32") 
            
            # posicion
            union.setPos( pos1 )
            
            # escala
            union.setScale(Escalas['m-px'] * 3 /32)
            
            union.setDepthTest(False)
            union.setBin("fixed", prioridad)
            union.reparentTo(parent)
            self.modelos.append(union)
                        
        # end-for
        
        # ULTIMO NODO DE UNION
        union = loader.loadModel("modelos/circulo32") 
        
        # posicion
        union.setPos( pos2 )
        
        # escala
        union.setScale(Escalas['m-px'] * 3 /32)
        
        union.setDepthTest(False)
        union.setBin("fixed", prioridad)
        union.reparentTo(parent)
        self.modelos.append(union)

        for nd in self.modelos:
            self.setTextura(nd,'peatonal', uScale=0.01, vScale=0.01)

#    def setTextura(self,seleccion):
#        
#        # carga una textura
#        textura = loader.loadTexture('texturas/'+ seleccion)
#        textura.setWrapU(Texture.WMMirror)
#        textura.setWrapV(Texture.WMMirror)
#        #textura.setBorderColor(VBase4(1, 1, 1, 0.0))
#        
#        for modelo in self.modelos:
#            
#            # crea las propiedades de la textura
#            stage = TextureStage('ts')
#            # estable la coordianacion de los vertices u,v
#            stage.setTexcoordName("coord")
#            # establce la textura(propiedades, textura, 1 idica prioridad sobre la textura anterior)
#            modelo.setTexture(stage, textura, 1)
#            # realiza la coordinacion de vertices
#            modelo.setTexGen(stage, TexGenAttrib.MWorldPosition)
#            #establece la escala de la textura
#            modelo.setTexScale(stage,Escalas['m-px']*2/64,Escalas['m-px']*2/64)
#            # desplaza la textura sobre el modelo
#            #modelo.setTexPos(stage,-0.39,-0.9,0)
#            # rota la textura
#            #modelo.setTexRotate(stage,-15)
        

class Plaza(Figura):
    
    def __init__(self,way,parent):
        
        self.modelos = []
        
        # obtiene la listade referencias de way
        referencias = way.getListaRefs()
        # envia todas las referencias menos la ultima, la primera es igual a la ultima
        # y no se puede enviar a Triangulator
        nodo = self.Triangular(referencias,parent)
        self.setTextura(nodo,'plaza', uScale=0.01, vScale=0.01)
        
        prioridad = 40
        nodo.setDepthTest(False)
        nodo.setBin("fixed", prioridad)
        nodo.reparentTo(parent)
        
        self.modelos.append(nodo)
        
        if way.getTagValue('name') is not None:
            self.setNombreIcono(way.getTagValue('name'), self.getCentroide(way), \
                                Escalas['m-px'] * 8, parent)
        
        self.setLineaTierra(self.getCentroide(way), Escalas['m-px'] * 8, parent)

class Garden(Figura):
    
    def __init__(self,way,parent):
        
        self.modelos = []
        
        nodo = self.Triangular(way.getListaRefs(),parent)
        
        prioridad = 40
        
        nodo.setDepthTest(False)
        nodo.setBin("fixed", prioridad)
        nodo.reparentTo(parent)
        self.setTextura(nodo, 'verde')
        
        self.modelos.append(nodo)
        
        if way.getTagValue('name') is not None:
            self.setNombreIcono(way.getTagValue('name'), self.getCentroide(way), \
                                Escalas['m-px'] * 8, parent)
           
            self.setLineaTierra(self.getCentroide(way), Escalas['m-px'] * 8, parent)

class Amenity(Figura):
    
    def __init__(self,way,parent):
        
        self.modelos = []
        
        
        if way.geTagValue('area') == 'yes':
            # envia todas las referencias menos la ultima, la primera es igual a la ultima
            # y no se puede enviar a Triangulator
            nodo = self.Triangular(way.getListaRefs(),parent)
        
            prioridad = 40
            
            nodo.setDepthTest(False)
            nodo.setBin("fixed", prioridad)
            nodo.reparentTo(parent)
            self.modelos.append(nodo)
            
            self.setTextura('salmon',128,128)
                
            nodo.reparentTo(parent)   

        else:
            nodo = OnscreenImagen(image='Icon-amenity_university.png',parent=parent)


 

#class Highway(Figura):
#    
#    def __init__(self,way,mapa,parent):
#        
#        # lista de NodePaths utilizados
#        self.modelos = []
#        
#        # determina el tipo de highway
#        textura = 'gris.png'
#        tags = way.getTags()
#        
#        if tags['highway'] == "primary":
#            #textura = 'carretera.png'
#            prioridad = 40
#            self.CrearHighway(textura,prioridad,way,mapa,parent)
#        else:
#            #textura = 'piedras.png'
#            prioridad = 40
#            self.CrearHighway(textura,prioridad,way,mapa,parent)
#        
#            
#    def CrearHighway(self,textura,prioridad,way,mapa,parent):
#
##        nrefs = way.getListaRefs().__len__()
##        refs = way.getListaRefs()
##        oneway = way.getTagValue('name')
##
##        # es un way formado por dos nodos(un único tramo de carretera)
##        if nrefs == 2:
##            
##            # obtiene las posiciones en el mapa de los nodos que componen way
##            pos1 = mapa.getNodoPosById(refs[0])
##            pos2 = mapa.getNodoPosById(refs[1])
##            
##            # solo hay un carril
##            if oneway == 'yes':
##                # modifica las posiciones
##                pos1,pos2 = self.CalcularPosicionesCarril(pos1, pos2)
##        
##            puente = loader.loadModel("modelos/carretera32")
##            
##            self.Puente(puente, pos1.getX(), pos1.getY(), pos,getX(), pos2.getY(), oneway)
##        
##            puente.setDepthTest(False)
##            puente.setBin("fixed", prioridad)
##            puente.reparentTo(parent)
##            self.modelos.append(puente)
##        
##            # nodo de union de inicio del carril
##            union = union = loader.loadModel("modelos/circulo32")
##            union.setPos(pos1)
##            
##            if oneway == 'yes':
##                escala = Escalas['m-px'] * 3.0 /32.0
##            else:
##                escala = Escalas['m-px'] * 6.0 /32.0
##            union.setScale(escala)
##            
##            self.modelos.append(union)
##            
##            # nodo de union del final del carril
##            union = union = loader.loadModel("modelos/circulo32")
##            union.setPos(pos2)
##            
##            if oneway == 'yes':
##                escala = Escalas['m-px'] * 3.0 /32.0
##            else:
##                escala = Escalas['m-px'] * 6.0 /32.0
##            union.setScale(escala)
##            
##            self.modelos.append(union)
##            
##        if nrefs >3:
##
##
##            # ref1: todos menos el ultimo
##            # ref2: todos menos el primero
##            
##            for ref in range(way.getListaRefs()[:-1].__len__()):
##                
##                ref1 = ref
##                ref2 = ref+1
##                # incluye el nuevo nodepath la final de la lista
##                modelo = loader.loadModel("modelos/carretera32")
##                
##                oneway = 'no'
##                if way.has_tag('oneway'):
##                    oneway = 'yes'
##                    #print oneway
##    
##                    # carretera
##                    tramo = self.CalcularPosicionesCarril(mapa.getNodoPosById(way.getListaRefs()[ref1]),
##                                                              mapa.getNodoPosById(way.getListaRefs()[ref2]))
##                    pos1 = tramo[0]
##                    pos2 = tramo[1]
##                    self.Puente(modelo, pos1[0], pos1[1], pos2[0], pos2[1], oneway)
##                
##                    # nodo de union
##                    union = loader.loadModel("modelos/circulo32")
##                    
##                    union.setX( pos1[0] )
##                    union.setY( pos1[1] )
##            
##                    escala = Escalas['m-px'] * 3.0 /32.0
##                    union.setScale(escala)
##                    
##                    union.setBin("fixed", prioridad+1)
##                    union.setDepthTest(False)
##                    modelo.setDepthWrite(False)
##                    union.reparentTo(parent)
##                    self.modelos.append(union)
##                
##                    union = loader.loadModel("modelos/circulo32")
##                    
##                    union.setX( pos2[0] )
##                    union.setY( pos2[1] )
##            
##                    escala = Escalas['m-px'] * 3.0 /32.0
##                    union.setScale(escala)
##                    
##                    union.setBin("fixed", prioridad+1)
##                    union.setDepthTest(False)
##                    modelo.setDepthWrite(False)
##                    union.reparentTo(parent)
##                    self.modelos.append(union)
##                
##                else:
##                    
##                    self.Puente(modelo,
##                                 mapa.getNodoXById( way.getListaRefs()[ref1] ),
##                                 mapa.getNodoYById( way.getListaRefs()[ref1] ),
##                                 mapa.getNodoXById( way.getListaRefs()[ref2] ),
##                                 mapa.getNodoYById( way.getListaRefs()[ref2] ),
##                                 oneway)
##                    
##                    
##               
##                    # nodo de union
##                    union = loader.loadModel("modelos/circulo32")
##                    
##                    escala = Escalas['m-px'] * 6.0 /32.0
##                    union.setScale(escala)
##                    
##                    union.setX( mapa.getNodoXById(way.getListaRefs()[ref1]) )
##                    union.setY( mapa.getNodoYById(way.getListaRefs()[ref1]) )
##                
##                    union.setBin("fixed", prioridad+1)
##                    union.setDepthTest(False)
##                    modelo.setDepthWrite(False)
##                    union.reparentTo(parent)
##                    self.modelos.append(union)
##                
##                
##                # el modelo de trabajo siempre esta al final de la lista 
##                modelo.setDepthTest(False)
##                modelo.setBin("fixed", prioridad)
##                modelo.reparentTo(parent)
##                self.modelos.append(modelo)
##                    
##            
##            # ultimo nodo de union
##            if oneway == 'no':
##                
##                union = loader.loadModel("modelos/circulo32")
##                
##                union.setX( mapa.getNodoXById(way.getListaRefs()[ref2]) )
##                union.setY( mapa.getNodoYById(way.getListaRefs()[ref2]) )
##                
##                escala = Escalas['m-px'] * 6.0 /32.0
##                union.setScale(escala)
##                
##                union.setBin("fixed", prioridad+1)
##                union.setDepthTest(False)
##                union.setDepthWrite(False)
##                union.reparentTo(parent)
##                self.modelos.append(union)
#        
########################################################        
########################################################
#
##
##        # ref1: todos menos el ultimo
##        # ref2: todos menos el primero        
##        for ref in range(way.getListaRefs()[:-1].__len__()):
##            
##            ref1 = ref
##            ref2 = ref+1
##            # incluye el nuevo nodepath la final de la lista
##            modelo = loader.loadModel("modelos/carretera32")
##            
##            oneway = 'no'
##            if way.has_tag('oneway'):
##                oneway = 'yes'
##                #print oneway
##            self.Puente(modelo,
##                         mapa.getNodoXById( way.getListaRefs()[ref1] ),
##                         mapa.getNodoYById( way.getListaRefs()[ref1] ),
##                         mapa.getNodoXById( way.getListaRefs()[ref2] ),
##                         mapa.getNodoYById( way.getListaRefs()[ref2] ),
##                         oneway)
##            
##            # el modelo de trabajo siempre esta al final de la lista 
##            modelo.setDepthTest(False)
##            modelo.setBin("fixed", prioridad)
##            modelo.reparentTo(parent)
##            self.modelos.append(modelo)
##           
##        # nodos de union    
##        for ref in way.getListaRefs():
##            modelo = loader.loadModel("modelos/circulo32") 
##            modelo.setX( mapa.getNodoXById(ref) )
##            modelo.setY( mapa.getNodoYById(ref) )
##            
##            escala = Escalas['m-px'] * 6.0 /32.0
##            if oneway == 'yes':
##                escala = escala/2
##            modelo.setScale(escala)
##            
##            modelo.setDepthTest(False)
##            modelo.setBin("fixed", prioridad+1)
##        
##            modelo.setDepthWrite(False)
##
##            modelo.reparentTo(parent)
##            self.modelos.append(modelo)
#                 
#############################################
############################################
#
#        # ref1: todos menos el ultimo
#        # ref2: todos menos el primero        
#        for ref in range(way.getListaRefs()[:-1].__len__()):
#            
#            ref1 = ref
#            ref2 = ref+1
#            
#            
#            way_oneway = 'no'
#            if way.has_tag('oneway'):
#                way_oneway = 'yes'
#                #print oneway
#                
#            
#            pos1 = mapa.getNodoPosById( way.getListaRefs()[ref1] )
#            pos2 = mapa.getNodoPosById( way.getListaRefs()[ref2] )
#            
#            """
#            la carretera se unen siempre por el centro excepto cuando se une una carretera
#            de un solo carrill a un nodo donde hay carreteras de dos carriles
#            """ 
#            
#            if way_oneway == 'yes': 
#            
#                nodo1_oneway = mapa.getOneWayByNodo(way.getListaRefs()[ref1])
#                nodo2_oneway = mapa.getOneWayByNodo(way.getListaRefs()[ref2])
#                
#                if nodo1_oneway == 'no' or nodo2_oneway == 'no':
#                    tramo = self.CalcularPosicionesCarril(pos1 ,pos2)
#                    
#                    if nodo1_oneway == 'no':
#                        pos1 = tramo[0]
#                        
#                    if nodo2_oneway == 'no':
#                        pos2 = tramo[1]
#                
#                modelo = loader.loadModel("modelos/carretera32")
#            
#            else:
#                modelo = loader.loadModel("modelos/carretera2carril32")
#                
#            self.Puente(modelo,pos1[0],pos1[1],pos2[0],pos2[1],way_oneway)
#            
#            # el modelo de trabajo siempre esta al final de la lista 
#            modelo.setDepthTest(False)
#            modelo.setBin("fixed", prioridad)
#            modelo.reparentTo(parent)
#            self.modelos.append(modelo)
#           
#            # NODO DE UNION
#            union = loader.loadModel("modelos/circulo32") 
#            union.setPos( pos1 )
#            
#            escala = Escalas['m-px'] * 6.0 /32.0
#            if way_oneway == 'yes':
#                escala = escala/2
#            union.setScale(escala)
#            
#            union.setDepthTest(False)
#            union.setBin("fixed", prioridad+1)
#        
#            union.setDepthWrite(False)
#
#            union.reparentTo(parent)
#            self.modelos.append(union)
#
#        # NODO DE UNION
#        union = loader.loadModel("modelos/circulo32") 
#        union.setPos( pos2 )
#        
#        escala = Escalas['m-px'] * 6.0 /32.0
#        if way_oneway == 'yes':
#            escala = escala/2
#        union.setScale(escala)
#        
#        union.setDepthTest(False)
#        union.setBin("fixed", prioridad+1)
#    
#        union.setDepthWrite(False)
#
#        union.reparentTo(parent)
#        self.modelos.append(union)
#        
#        self.CrearAcera(textura, prioridad, way, mapa, parent)
#        
#
#    def CrearAcera(self,textura,prioridad,way,mapa,parent):
#        
#        # ref1: todos menos el ultimo
#        # ref2: todos menos el primero        
#        for ref in range(way.getListaRefs()[:-1].__len__()):
#            
#            ref1 = ref
#            ref2 = ref+1
#            # incluye el nuevo nodepath la final de la lista
#            modelo = loader.loadModel("modelos/acera32")
#         
#            
#            modelo.detachNode()
#            
#            way_oneway = 'no'
#            if way.has_tag('oneway'):
#                way_oneway = 'yes'
#                #print oneway
#                
#            
#            pos1 = mapa.getNodoPosById( way.getListaRefs()[ref1] )
#            pos2 = mapa.getNodoPosById( way.getListaRefs()[ref2] )
#            
#            """
#            la carretera se unen siempre por el centro excepto cuando se une una carretera
#            de un solo carrill a un nodo donde hay carreteras de dos carriles
#            """ 
#            
#            if way_oneway == 'yes': 
#            
#                nodo1_oneway = mapa.getOneWayByNodo(way.getListaRefs()[ref1])
#                nodo2_oneway = mapa.getOneWayByNodo(way.getListaRefs()[ref2])
#                
#                if nodo1_oneway == 'no' or nodo2_oneway == 'no':
#                    tramo = self.CalcularPosicionesCarril(pos1 ,pos2)
#                    
#                    if nodo1_oneway == 'no':
#                        pos1 = tramo[0]
#                        
#                    if nodo2_oneway == 'no':
#                        pos2 = tramo[1]
#                        
#            
#            self.Puente(modelo,pos1[0],pos1[1],pos2[0],pos2[1],way_oneway)
#            
#            if way_oneway == 'yes':
#                modelo.setSy(Escalas['m-px'] * 6.5 /32.0 * 0.5)
#            else:
#                modelo.setSy(Escalas['m-px'] * 6.5 /32.0)
#            
#            # el modelo de trabajo siempre esta al final de la lista 
#            modelo.setDepthTest(False)
#            modelo.setBin("fixed", prioridad-2)
#            modelo.reparentTo(parent)
#            self.modelos.append(modelo)
#           
#            # NODO DE UNION
#            union = loader.loadModel("modelos/acera_circulo32") 
#            union.setPos( pos1 )
#            
#            escala = Escalas['m-px'] * 6.5 /32.0
#            if way_oneway == 'yes':
#                escala = escala/2
#            union.setScale(escala)
#            
#            union.setDepthTest(False)
#            union.setBin("fixed", prioridad-2)
#        
#            union.setDepthWrite(False)
#
#            union.reparentTo(parent)
#            self.modelos.append(union)
#
#        # NODO DE UNION
#        union = loader.loadModel("modelos/acera_circulo32") 
#        union.setPos( pos2 )
#        
#        escala = Escalas['m-px'] * 6.5 /32.0
#        if way_oneway == 'yes':
#            escala = escala/2
#        union.setScale(escala)
#        
#        union.setDepthTest(False)
#        union.setBin("fixed", prioridad-2)
#    
#        union.setDepthWrite(False)
#
#        union.reparentTo(parent)
#        self.modelos.append(union)
#
#    # recibe las posiciones de los tres nodos que componen
#    # los dos tramos de caretera, halla la interseccion de las
#    # dos rectas de los dos carriles
#    def CalcularCruceCarriles(self,pos1,pos2,pos3):
#
#        #la rectas que marcan los carriles no pasan
#        # por le centro de la carretera, hay que calcular
#        # el desplazamiento, pos2 y pos3 estan asociados
#        # al mimso nodo pero a rectas diferentes
#        
#        # puntos de referencia para la primera recta        
#        tramo1 = self.CalcularPosicionesCarriles(pos1, pos2)
#        
#        # puntos de referencia para la segunda recta
#        tramo2 = self.CalcularPosicionesCarriles(pos2, pos3)
#        
#        # tupla posicion de inicio del carril
#        pos1 = tramo1[0]
#        # tupla posicion de final del carril
#        pos2 = tramo1[1]
#                
#        pos3 = tramo2[0] 
#        pos4 = tramo2[1]
#        
#            
#        # si los dos primero puntos estan en la misma x
#        # calcula para que y pasa por x la segunda recta
#        if abs(pos1[0]-pos2[0])<0.001 and abs(pos3[0]-pos4[0])<0.001:
#            # al final de su carril
#            x = pos2[0]
#            y = pos2[1]
#            
#        elif abs(pos1[0]-pos2[0])<0.001:
#            x = pos1[0]
#            
#            m2 = (pos3[1]-pos4[1])/(pos3[0]-pos4[0])
#            b2 = pos3[1] - m2*pos3[0]
#            y = m2*x + b2
#            
#        # si el segundo y tercer punto esta en la misma x
#        # calcual para que y para por x la primera recta
#        elif abs(pos3[0]-pos4[0])<0.001:
#            x = pos3[0]
#            
#            m1 = (pos1[1]-pos2[1])/(pos1[0]-pos2[0])
#            b1 = pos1[1] - m1*pos1[0]
#            y = m1*x + b1
#            
#            
#        # en otro caso calcula ambas recta y su punto de corte
#        else:
#            # m = (y1 - y2)/(x1 - x2)
#            m1 = (pos1[1]-pos2[1])/(pos1[0]-pos2[0])
#            m2 = (pos3[1]-pos4[1])/(pos3[0]-pos4[0])
#            
#            if abs(m1-m2)<0.001:
#                # las rectar tiene la misma direccion
#                # coinciden al final e inicio de carril
#                # los dos tramos estan en linea
#                x = pos2[0]
#                y = pos2[1]
#            
#            else:
#                # b = y - mx
#                b1 = pos1[1] - m1*pos1[0]
#                b2 = pos3[1] - m2*pos3[0]
#                
#                x = (b2-b1)/(m1-m2)
#            
#                y = m1*x + b1
#     
#        return VBase3(x,y,0)
#
#    # calcula la posicion de partida de cada carril
#    # formado para cada par de nodos
#    def CalcularPosicionesCarril(self,pos1,pos2):
#        
#        # angulo desde pos1 a pos 2
#        """
#        vx = pos2[0] - pos1[0]
#        vy = pos2[1] - pos1[1]
#        """
#        vxy = pos2 - pos1
#        
#        if(vxy.getX() == 0):
#            anguloDegrees = 90
#            #self.setHeading(90)
#        elif(vxy.getY() == 0):
#            anguloDegrees = 0
#            #self.setHeading(0)
#        else:
#            angulo = abs( vxy.getY() / vxy.getX() ) 
#            anguloRad = atan(angulo)
#            anguloDegrees = degrees(anguloRad)
#        # el calculo de los angulos se puede trasladar a cuando
#        # a cuando se direge a su nuevo destino
#        if vxy.getX()<=0 and vxy.getY()>0: # segundo cuadrante
#            anguloDegrees = 180 - anguloDegrees
#        elif vxy.getX()<=0 and vxy.getY()<=0: # tercer cuadrante
#            anguloDegrees = anguloDegrees + 180
#        elif vxy.getX()>0 and vxy.getY()<=0: # cuarto cuadrante
#            anguloDegrees = anguloDegrees * (-1)
#        
#        # desplazamiento desde pos1
#        alpha = anguloDegrees - 90
#        betaRadians = radians(alpha)
#        
#        desplaz_carril = Escalas['m-px']*1.2
#        
#        pos1 =  ( VBase3(pos1[0]+cos(betaRadians)*desplaz_carril , pos1[1]+sin(betaRadians)*desplaz_carril,0) )
#        """
#        pos1.addX(cos(betaRadians)*3.2)
#        pos1.addY(sin(betaRadians)*3.2)
#        """
#        
#        #desplazamiento desde pos2
#        #alpha = anguloDegrees + 90
#        #omegaRadians = radians(alpha)
#        #pos2 =  ( (pos2[0]+cos(omegaRadians)*3.2 , pos2[1]+sin(omegaRadians)*3.2) )
#        
#        pos2 =  ( VBase3(pos2[0]+cos(betaRadians)*desplaz_carril , pos2[1]+sin(betaRadians)*desplaz_carril,0) )
#        """
#        pos2.addX(cos(betaRadians)*3.2)
#        pos2.addY(sin(betaRadians)*3.2)
#        """
#        return (pos1,pos2)
#  
#    #situa el objeto puente entre dos nodos del escenario
#    # parametro: NodePath, posiciones (x,y) en pixeles

class Area(Figura):
    
    def __init__(self,way,parent,imagen):

        self.modelos = []
        
        area = self.Triangular(way.getListaRefs(),parent)
        self.setTextura(area, 'textura1', uScale=0.01, vScale=0.01)
        area.setBin("fixed",30)
        area.setDepthTest(False)
        
        altura = Escalas['m-px'] * 7
    
        self.setAvatar(way.getTagValue('name'),
                       imagen,
                       self.getCentroide(way),
                       altura,
                       parent)

class Records:
    
    def __init__(self):
        
        try:
            # abre el fichero de recods
            self.fileRecords = file('records.dat')
            
            # carga los records anteriores
            self.records = pickle.load(self.fileRecords)
             
            # cierra el fichero
            self.fileRecords.close()
            
            print 'Records: archivo de records leido'
            
        except :
            
            # si el fichero no existe los crea
            self.fileRecords = file('records.dat','w')
            # cierra el fichero
            self.fileRecords.close()
            # crea la variable de recods vacia
            self.records = {}
            
            print 'Records: archivo de records nuevo creado'
            
    
    def RecuperarRecord(self,way_origen,way_destino):
    
        print 'Record:RecuperarRecord'
        
        if self.records.has_key(way_origen + way_destino):
            return self.records[way_origen + way_destino]
        else:
            return None
         
    def GuardarRecord(self,way_origen,way_destino,tiempo):
        
        print 'Record:GuardarRecord'
        
        # si habia un registro anterior
        if self.records.has_key(way_origen + way_destino):
            
            # si el nuevo tiempo es menor lo sustituye
            tiempo_guardado = self.RecuperarRecord(way_origen, way_destino)
            if tiempo < tiempo_guardado:
                self.records[way_origen + way_destino] = tiempo
                
                # guarda los registros
                self.GuardarAFile()
        
        # no hay regustros anteriores y lo almacena
        else:
            self.records[way_origen + way_destino] = tiempo
            self.GuardarAFile()
    
    def GuardarAFile(self):
        
        print 'Records:GuardarAFile'
        
        # si el fichero no existe los crea
        self.fileRecords = file('records.dat','w')
        # guarda los datos
        pickle.dump(self.records,self.fileRecords,2)
        # cierra el fichero
        self.fileRecords.close()
    
    
"""
Flecha de orientacion utilizada en el juego
"""
class Flecha:
    
    def __init__(self,nodo,destino):
        
        self.flecha = OnscreenImage(image="imagenes/flecha3.png",parent=nodo)
        self.flecha.reparentTo(nodo)
        self.flecha.setScale(Escalas['m-px']*1.5)
        self.flecha.setHpr(0,-90,0)
        
        self.flecha.setTransparency(TransparencyAttrib.MAlpha)
        self.flecha.setColor(1,1,1,0.8)
        
        prioridad = 110
        self.flecha.setBin("fixed", prioridad)
      
        self.destino
            
    def setPos(self,pos):
        self.flecha.setPos(pos)
    
    # recoloca el nodo hace el que se dirige  el vehiculo 
    def setDestino(self,pos):
        self.destino.setPos(pos) 
    
    # reorienta la flecha
    def Orientar(self):
        self.flecha.lookAt(self.destino)
        self.flecha.setP(-90)


# Muestra el camino creado por el algoritmo de Djkstra     
class Camino:
    
    def __init__(self,NPparent):
        
        self.parent = NPparent
        
        # lista de flechas creadas para formar el camino
        self.listaFlechas = []
    
        # nodo auxiliar para poder orientar la flecha
        self.destino = NodePath('orientacion')
    
    def MostrarCamino(self,lista):
        
        print 'Camino:MostrarCamino'
        
        i = 0
        while i < lista.__len__()-1:
            flecha = OnscreenImage(image="imagenes/flecha_amarilla.png",parent=self.parent)
            flecha.setScale(Escalas['m-px']*3)
            flecha.setHpr(0,-90,0)
            flecha.setPos(lista[i].getPos())
        
            flecha.setTransparency(TransparencyAttrib.MAlpha)
            flecha.setColor(1,1,1,0.8)
        
            prioridad = 105
            flecha.setBin("fixed", prioridad)
            flecha.setDepthTest(False)

            self.listaFlechas.append(flecha)
            self.Orientar(lista[i+1].getPos(),i)
            
            i = i+1
    
            
        flecha = OnscreenImage(image="imagenes/diana.png",parent=self.parent)
        flecha.setScale(Escalas['m-px']*3)
        flecha.setHpr(0,-90,0)
        flecha.setPos(lista[i].getPos())
        flecha.setTransparency(TransparencyAttrib.MAlpha)
        flecha.setColor(1,1,1,0.8)
    
        prioridad = 120
        flecha.setBin("fixed", prioridad)
        flecha.setDepthTest(False)
  
        self.listaFlechas.append(flecha)
    
    # reorienta la flecha
    def Orientar(self,pos,i):
        self.destino.setPos(pos)
        self.listaFlechas[i].lookAt(self.destino)
        self.listaFlechas[i].setP(-90)
    
#    def CalcularHeading(self,pos1,pos2):
#        
#        vx = pos2[0] - pos1[0]
#        vy = pos2[1] - pos1[1]
#        
#        if(vx == 0):
#            anguloDegrees = 90
#            #self.setHeading(90)
#        elif(vy == 0):
#            anguloDegrees = 0
#            #self.setHeading(0)
#        else:
#            angulo = abs( vy / vx ) 
#            anguloRad = atan(angulo)
#            anguloDegrees = degrees(anguloRad)
#        # el calculo de los angulos se puede trasladar a cuando
#        # a cuando se direge a su nuevo destino
#        if vx<0 and vy>0: # segundo cuadrante
#            anguloDegrees = 180 - anguloDegrees
#        elif vx<0 and vy<0: # tercer cuadrante
#            anguloDegrees = anguloDegrees + 180
#        elif vx>0 and vy<0: # cuarto cuadrante
#            anguloDegrees = anguloDegrees * (-1)
#        
#        self.flecha.setH(anguloDegrees-90)
    
    def EliminarCamino(self):
        
        # elimina el camino
        while self.listaFlechas:
            self.listaFlechas[-1].removeNode()
            self.listaFlechas.pop()
            
    def __del__(self):

        # elimina camino
        while self.listaFlechas:
            self.listaFlechas[-1].removeNode()
            self.listaFlechas.pop()


class Cielo():
    
    def __init__(self,parent):
        
        self.cielo = loader.loadModel('modelos/sunny')
        self.cielo.setScale(Escalas['m-px']*125/256)    
        self.cielo.setBin('background', 10)
        self.cielo.setDepthWrite(0)
#        self.cielo.setLightOff()
        
        self.cielo.reparentTo(parent)
        
    def setPos(self,pos):
        self.cielo.setPos(pos) 
        
class Suelo():
    
    def __init__(self,parent):
        
        self.suelo = loader.loadModel("modelos/highway")
        
#        type(self.suelo.node())        
#        # coloca la figura en el centro del mapa
#        x = (mapa.getBoundsMinX() + mapa.getBoundsMaxX() ) / 2
#        y = (mapa.getBoundsMinY() + mapa.getBoundsMaxY() ) / 2
#        self.modelos[0].setPos(x,y,0)
#        
#        x_escala = mapa.getBoundsMaxX()
#        y_escala = mapa.getBoundsMaxY()
#        # dividido por 100 que es el tamano en pixeles de la figura
#        print 'escala suelo', x_escala,y_escala

        dist = base.camLens.getFar()
        self.suelo.setScale(dist*2/100,dist*2/100,1)
        
        # TEXTURA
        
        # carga una textura
        textura = loader.loadTexture('texturas/marron.png')
        textura.setWrapU(Texture.WMMirror)
        textura.setWrapV(Texture.WMMirror)
        #textura.setBorderColor(VBase4(1, 1, 1, 0.0))
            
        # crea las propiedades de la textura
        stage = TextureStage('ts')
        # estable la coordianacion de los vertices u,v
        stage.setTexcoordName("coord")
        # establce la textura(propiedades, textura, 1 idica prioridad sobre la textura anterior)
        self.suelo.setTexture(stage, textura, 1)
        # realiza la coordinacion de vertices
        self.suelo.setTexGen(stage, TexGenAttrib.MWorldPosition)
        #establece la escala de la textura
        self.suelo.setTexScale(stage,0.5,0.5)
        # desplaza la textura sobre el modelo
        #modelo.setTexPos(stage,-0.39,-0.9,0)
        # rota la textura
        #modelo.setTexRotate(stage,-15)

        
        # lower value are drawn first.
        self.suelo.setBin("fixed", 30)
        self.suelo.setDepthTest(False)
        
        self.suelo.reparentTo(parent)
        
    def setPos(self,p):
        p.setZ(0)
        self.suelo.setPos(p)

class MapaMini:
    
    def __init__(self,nodo):
        
        print "MapaMini"
        
        self.parent = nodo
        
        # crea la region para la nueva camara
        # left right bottom top [0,1]
        self.dr1 = base.win.makeDisplayRegion(0.7, 1, 0.7, 1)
       
        # crea la camara
        self.camara = Camera('cam')
        lens = OrthographicLens()
        lens.setFilmSize(1000,1000)  # Or whatever is appropriate for your scene
        lens.setFar(600)
        self.camara.setLens(lens)
        
        # crea un nodepath a partir de la camara
        self.camaraNP = NodePath(self.camara) 
        # configura la camara para la nueva region
        self.dr1.setCamera(self.camaraNP)
        # establece el orden de mostrado
        self.dr1.setSort(10)
        
        # introduce la camara en la escena
        # inicialmente la camara no esta unida al escenario
        #self.camaraNP.reparentTo(nodo)
        # posicion la camara para la vista aerea
        self.camaraNP.setPos(0,0,500)
        self.camaraNP.setP(-90)

        
        # crea la segunda region sobre la anterior
        self.dr2 = base.win.makeDisplayRegion(0.7,1,0.7,1)
        # prioridad de renderizado
        self.dr2.setSort(20)

        # crea la camara2d
        self.myCamera2d = NodePath(Camera('myCam2d'))
        # establece las propiedades de la optica
        lens = OrthographicLens()
        lens.setFilmSize(2, 2)
        lens.setNearFar(-1000, 1000)
        self.myCamera2d.node().setLens(lens)
        
        # crea un nodo generico para el nuevo escenario
        self.myRender2d = NodePath('myRender2d')
        self.myRender2d.setDepthTest(False)
        self.myRender2d.setDepthWrite(False)
        # asocia la camara al nuevo nodo
        self.myCamera2d.reparentTo(self.myRender2d)
        # confugra la camara para la nueva displayRegion
        self.dr2.setCamera(self.myCamera2d)

        # radar del vehiculo
        self.brujula = OnscreenImage(image='imagenes/punta2.png',parent=None,scale=0.1)
        # hace transparente la flecha
        self.brujula.setTransparency(TransparencyAttrib.MAlpha)
        
        # marco del mapa
        self.marco = OnscreenImage(image='imagenes/marco2.png',parent=None,scale=1)
        # hace transparente la flecha
        self.marco.setTransparency(TransparencyAttrib.MAlpha)
        
        
        self.Ocultar()
        
    def setPos(self,pos):
        self.camaraNP.setPos(pos)
        """
        self.camaraNP.setX(pos[0])
        self.camaraNP.setY(pos[1])
        """
        
    def setRoll(self,r):
        # la brujula se muecve en sentido horario
        # tiene un angulo por defecto de 90 grados
        self.brujula.setR(360-(r-90))
    
    """
    def reparentTo(self,nodo):
        self.myCamera2d.reparentTo(nodo)
    
    """
     
    def DisminuirZoom(self):
        # obtiene el tamaño de la lente
        zoom = self.camaraNP.node().getLens().getFilmSize()
        zoom = zoom + VBase2(100,100)
        # establece el nuevo tamaño
        self.camaraNP.node().getLens().setFilmSize(zoom)
        
    def AumentarZoom(self):
        # obtiene el tamaño de la lente
        zoom = self.camaraNP.node().getLens().getFilmSize()
        if zoom.getX()>100:
            zoom = zoom - VBase2(100,100)
            self.camaraNP.node().getLens().setFilmSize(zoom)
     
    def Mostrar(self):
        # dr1
        self.camaraNP.reparentTo(self.parent)
        # dr2
        self.brujula.reparentTo(self.myRender2d)
        self.marco.reparentTo(self.myRender2d)
        
    def Ocultar(self):
        # dr1
        self.camaraNP.detachNode()
        # dr2
        self.brujula.detachNode()
        self.marco.detachNode()
        
    def __del__(self):
        
        #self.camaraNP.removeNode()
        self.myCamera2d.removeNode()
        self.myRender2d.removeNode()
        
        self.brujula.removeNode()
        self.marco.removeNode()
    
class Escenario: 
    
    def __init__(self):
        
        print "Escenario"
        
        # NodePath del menu a patir del cual cuelgan el resto de nodos
        self.NodoEscenario = NodePath('NodoEscenario')
        self.NodoEscenario2d = NodePath('NodoEscenario2D')
        
        # variable local del objeto Menu para los eventos
        self.EventosEscenario = DirectObject()

        # lista de elementos que aparecen en el escenario
        self.Figuras = []                    
    
        # FUENTES
        fontSansBold = loader.loadFont('fuentes/FreeSansBold.ttf')
        fontSerif = loader.loadFont('fuentes/FreeSerif.ttf')
    
        # MUSICA
        
        self.musica = loader.loadSfx('./sonidos/escenario.ogg')
        self.musica.setLoop(True)

    
        # INSTRUCCIONES
        
        # muestra el texto de instrucciones
        self.texto_ayuda = OnscreenText(text="Pulse 'F1' para mostra la ayuda.",
                                     align=TextNode.ALeft, pos =(-1.3,-0.95), scale=0.05,
                                     fg=(1,1,1,1),
                                     bg=(0,0,0,0.4),
                                     font = fontSansBold,
                                     parent=self.NodoEscenario2d)
        
        # VEHICULOS
        self.circulacion = Circulacion(self.NodoEscenario,self.NodoEscenario2d)
        
        # CUPULA DEL CIELO
        self.cielo = Cielo(self.NodoEscenario)
        
        # SUELO
        self.suelo = Suelo(self.NodoEscenario)
        
        # MAPA PEQUEÑO EN LA ESQUINA
        self.mapa_mini = MapaMini(self.NodoEscenario)
        
        # POSICION DE INICIO DEL JUGADOR
        self.MenuDireccionNodePath = NodePath('pos_inicio')
        
        self.textoMenuDireccion = OnscreenText(text='',
                                            align = TextNode.ACenter,
                                            font = fontSerif,
                                            pos = (0,0),
                                            fg = (1,1,1,1),
                                            bg = (0,0,0,0.4),
                                            parent = self.MenuDireccionNodePath,
                                            mayChange = True
                                            )
        
        # texto que acompaña a la caja de textp
        self.textoBuscarDir = OnscreenText(text='Escriba el nombre de la calle:',
                                            align = TextNode.ALeft,
                                            font = fontSerif,
                                            pos = (-1,-0.5),
                                            fg = (1,1,1,1),
                                            bg = (0,0,0,0.4),
                                            parent = self.MenuDireccionNodePath
                                            )
        # caja de texto
        self.DEntryBuscarDir = DirectEntry(text = "" , scale=0.08, pos = (-0.10,0,-0.5),
                                     initialText="", numLines = 1, frameColor=(0,0,0,0.4),
                                     text_fg = (1,1,1,1), command = self.setFocus,
                                     width=15, entryFont = fontSerif, parent=self.MenuDireccionNodePath)
        
        
        
         # lista de tuplas con los nombre/idWay de las callles para la poscion de inicio del jugador
        self.listaCalles = []
        # lista de calles que coinciden con la busqueda,tupla(nombre,listaRefs,way_id)
        self.listaCallesBusqueda = []
        # indice de la calle de inicio del jugador seleccionada
        self.indiceCalle =  0
        

        #LIMITES DEL JUGADOR EN EL MAPA
        
        # tamaño del circulo del mapa pintado alrededor del jugador
        self.tam_mapa = base.camLens.getFar()*2
        
        cNodeBounds = CollisionNode( 'mini_bounds' )
        cNodeBounds.addSolid(CollisionInvSphere(0, 0, 0, 0.85*self.tam_mapa))
        # mascara al azar para que no coincida
        cNodeBounds.setFromCollideMask(BitMask32(0x22))
        # mascara from del vehiculo de jugador
        cNodeBounds.setIntoCollideMask(BitMask32(0x02))
        self.cNodePathBoundsJugador = NodePath(cNodeBounds)
#        self.cNodePathBounds.show()
        self.cNodePathBoundsJugador.reparentTo(self.NodoEscenario)
        
        # nodo vacio, para evitar errores, el orginal se crea en Entrar()
        self.cNodePathBoundsMapa = NodePath('vacio')
        
        #self.mapa_mini.setRoll(self.jugador.getHeading())
    
        # CAMINO GENERADO POR EL ALGORITMO DE DJKSTRA
        self.camino = Camino(self.NodoEscenario)
    
        # texto para la notificacion de errores en el escenario
        self.textoError = OnscreenText(text='Se ha producido un error al cargar el escenario.',
                                       pos=(0,0,0),
                                       fg=(1,1,1,1),
                                       bg=(0,0,0,0.4),
                                       scale = 0.1,
                                       align = TextNode.ACenter,
                                       font = fontSerif,
                                       parent=None,
                                       mayChange=True)
        self.textoError.detachNode()
    
        # RECORDS
        
        # objeto de medicion de records
        self.records = Records()
        
        # texto para monstrar mensajes desde el menu records
        self.textoRecord = OnscreenText(text='',
                                       pos=(0,0,0),
                                       fg=(1,1,1,1),
                                       bg=(0,0,0,0.4),
                                       scale = 0.1,
                                       align = TextNode.ACenter,
                                       font = fontSerif,
                                       parent=None)
        self.textoRecord.detachNode()
        
        # texto para mostrar la direccion de inicio
        self.textoRecordInicio = OnscreenText(text='',
                                            align = TextNode.ACenter,
                                            font = fontSerif,
                                            pos = (0,-0.3),
                                            fg = (1,1,1,1),
                                            bg = (0,0,0,0.4),
                                            parent = None,
                                            mayChange = True
                                            )
        self.textoRecordInicio.detachNode()
        
        # texto para mostrar el tiempo transcurrido
        self.textoRecordTiempo = OnscreenText(text='',
                                     align=TextNode.ARight, pos =(1.3,-0.95), scale=0.05,
                                     fg=(1,1,1,1),
                                     bg=(0,0,0,0.4),
                                     parent=None
                                     )   
        self.textoRecordTiempo.detachNode()  
        
        # almacena los nodepath creados para la meta
        self.listaNodePathRecord = []
        
        # variable para medir el tiempo
        self.recordTiempo = 0
        
        #JUEGO
        
        # texto para monstrar mensajes desde el menu del juego
        self.textoJuego = OnscreenText(text='',
                                       pos=(0,0,0),
                                       fg=(1,1,1,1),
                                       bg=(0,0,0,0.4),
                                       scale = 0.1,
                                       align = TextNode.ACenter,
                                       font = fontSerif,
                                       parent=None)
        self.textoJuego.detachNode()
    
        # texto para mostrar el tiempo transcurrido
        self.textoJuegoTiempo = OnscreenText(text='',
                                     align=TextNode.ARight, pos =(1.3,-0.95), scale=0.05,
                                     fg=(1,1,1,1),
                                     bg=(0,0,0,0.4),
                                     parent=None
                                     )   
        self.textoJuegoTiempo.detachNode()  
        
        # variable para medir el tiempo
        self.tiempoJuego = 0
        
        # destino del juego
        self.juegoDestino = None
        
        # almacena los nodepath creados para la meta
        self.listaNodePathJuego = []
        
        # INSTRUCCIONES
        
        # texto para mostrar el control desde teclado
        self.textoControles = OnscreenText(text='Controles:\n\
Acelerar: flecha-arriba \n\
Frenar/detener: flecha-abajo \n\
Cambiar dirección: d \n\
Invertir sentido: r \n\
Buscar ruta/cancelar: control-f \n\
Ir a/cancelar: control-g \n\
Contrarelog/cancelar: control-h \n\
Juego encuentra la dirección/cancelar: control-j \n\
Para filtrar una dirección: shift-enter\n\
Cambiar dirección: flechas-arriba/abajo\n\
Salir: control-q\
',
                                            align = TextNode.ALeft,
                                            font = fontSerif,
                                            pos = (-0.7,0.3),
                                            fg = (1,1,1,1),
                                            bg = (0,0,0,0.9))
        self.textoControles.detachNode()
        
        
    # recibe de donde se obtendra el mapa y el nombre o las coordenadas
    def Entrar(self,seleccion,dato1,dato2,nodo_origen=None,nodo_destino=None):
           
        print 'Escenario:Entrar'      
        
        self.musica.play()
        
        # CARGA EL MAPA 
        
        # alamacena el tipo de mapa utilizado
        self.modoMapa = seleccion        
        
#        try:
        # establece el enlace con el documentos xml
        if seleccion == 'lista':
            mapa = Mapa(dato1,Escalas['gr-px'])
        else: # seleccion == 'http'
            mapa = Mapa("http",Escalas['gr-px'],float(dato1),float(dato2))
        print "Mapa creado"
            # recorreo todos los way que componen el mapa
#        except:
#            print 'Error. No se pudo cargar el nuevo mapa'
#            self.textoError.node().setText('Error. No se pudo cargar el mapa.')
#            self.textoError.reparentTo(self.NodoEscenario2d)
#            
#            # REPARENT
#            self.NodoEscenario.reparentTo()
#            self.NodoEscenario2d.reparentTo(aspect2d)
#            
#            return 0
#        
        self.textoError.detachNode()
    
        self.mapa = mapa
        
        # LIMITES DEL MAPA
        cNodeBounds = CollisionNode( 'bounds' )
        cNodeBounds.addSolid( CollisionPlane(Plane(Vec3(1, 0, 0), Point3(self.mapa.getBounds().getMinPos()) )) )
        cNodeBounds.addSolid( CollisionPlane(Plane(Vec3(0, 1, 0), Point3(self.mapa.getBounds().getMinPos()) )) )
        cNodeBounds.addSolid( CollisionPlane(Plane(Vec3(-1, 0, 0), Point3(self.mapa.getBounds().getMaxPos()) )) )
        cNodeBounds.addSolid( CollisionPlane(Plane(Vec3(0, -1, 0), Point3(self.mapa.getBounds().getMaxPos()) )) )
        cNodeBounds.setFromCollideMask(BitMask32(0x22))
        cNodeBounds.setIntoCollideMask(BitMask32(0x02))
        self.cNodePathBoundsMapa = self.NodoEscenario.attachNewNode(cNodeBounds)
#        self.cNodePathBounds.show() 
        
        # RECOPILACION DE LOS NOMBRES DE LAS CALLES
        
        # recorre la lista de carreteras
        for w in self.mapa.getListaCarreteras():
#            nombre = w.getTagValue('name')
            if w.has_tag('name') and w.getListaRefs().__len__() > 0:
                self.listaCalles.append(w)    
                
#         MOVIDO DE DIRECCION
#        self.DEntryBuscarDir.enterText('')
#        # para buscar las calles coincidentes
#        self.EventosEscenario.accept('shift-enter',self.BuscarDireccionSeleccionada)
#        
        
        # LISTA DE LAS POSICION DE INICIO DEL JUGADOR
        # si no hay direcciones disponible
        # si se especifica un nodo de origen de un mapa anterior
        if self.listaCalles.__len__() == 0 or nodo_origen is not None:
            self.ComenzarCirculacionAutomatico(nodo_origen, nodo_destino)
            # si no hay direcciones disponible se elegi una posicion al azar
        else:
            self.MenuDireccion()
            
        
        # REPARENT
        self.NodoEscenario.reparentTo(render)
        self.NodoEscenario2d.reparentTo(aspect2d)
        
        # MAPA MINI
        self.mapa_mini.Mostrar()
        
        # enfoca a lo alto del cielo
        base.camera.setP(90)
        
        # muestra y oculta el texto de controles
        self.EventosEscenario.accept('f1',self.MostrarControles)
        self.EventosEscenario.accept('f1-up',self.OcultarControles)
        

#        except:
#            print 'Escenario:Entrar, error al cargar el mapa'
#           
#            # REPARENT
#            self.NodoEscenario.reparentTo()
#            self.NodoEscenario2d.reparentTo(aspect2d)
#            
#            # MAPA MINI
#            self.mapa_mini.Mostrar()
#            
#            #SUELO
#            self.suelo.setPos(base.camera.getPos())
#            
#            # enfoca a lo alto del cielo
#            base.camera.setP(90)
#            
#            self.textoError.reparentTo(self.NodoEscenario)
#        
     
#        DIVISION EN SECCIONES
#
#        # el mapa se dividira en secciones
#        factor_seccion = Escalas['m-px']*1000
#        # tamaño en x del mapa
#        mapa_x = self.mapa.getBoundsMaxX()
#        mapa_y = self.mapa.getBoundsMaxY()
#        
##        print 'factor seccion', factor_seccion
##        print 'dimension mapa', mapa_x,mapa_y
#    
#        # número de secciones por eje
#        nseccion_x = int ( ceil(mapa_x / factor_seccion) )
#        nseccion_y = int ( ceil(mapa_x / factor_seccion) ) 
#        
##        print 'numero de secciones',nseccion_x,nseccion_y
#        
#        for i in range(nseccion_x):
#            for j in range(nseccion_y):
#                self.secciones[str(i)+str(j)] = NodePath(str(i)+str(j))
##        print self.secciones
#        
#        for way in mapa.getListaWays():
#           
#            # si es una carreteras
#            if w.has_tag('highway'):
#                # obtiene su nombre
#                nombre = w.getTagValue('name')
#                if nombre and w.getListaRefs().__len__() > 0:
#                    self.listaCalles.append( (nombre,w.getListaRefs()) )
##                else:
##                    refs = w.getListaRefs()
##                    if refs.__len__() > 0:
##                        pos = self.mapa.getNodoPosById(refs[0])
##                        nombre = 'Sin nombre(coordenadas: ' + str( pos.getX() ) + ',' + str( pos.getY() ) + ')'
##                        self.listaCalles.append( (nombre,w.getId() ) )
#        
#           
#            # para cada lista de tags
#            tags = way.getTags()
#            
#            """
#            if tags.has_key('name'):
#                #print tag.getValue()
#                self.Figuras.append(Nombre(way,mapa,self.NodoEscenario))
#                pass 
#            """
#            # para que muestra los way id
##            self.Figuras.append(Nombre(way,mapa,self.NodoEscenario))
#            
#            # si way es del tipo hihgway
#            if tags.has_key('highway'):
#                #crea un higway y lo incluye en la lista para almacenarlo
#                self.Figuras.append( Highway(way,mapa,self.NodoEscenario) )
#                pass
#            """
#            if tags.has_key('leisure'):
#                # crea un leisure y lo incluye en la lista para almacenarlo
#                self.Figuras.append( Leisure(way,mapa,self.NodoEscenario) )
#                pass
#            
#            if tags.has_key('natural'):
#                #self.Figuras.append( Highway(way,mapa,self.NodoEscenario) )
#                pass 
#            """
        
            
        #funcion demovimiento de camaraa ejecutar con cada bucle del videojuego
        #taskMgr.add(self.Movimiento,"camara")
        
        #posicion inicial de la camara
        #base.camera.setPos(100,100,100)
        #base.camera.setHpr(0,0,0)
        """
        #controles
        self.EventosEscenario.accept("w", self.setKey, ["w",1])
        self.EventosEscenario.accept("w-up", self.setKey, ["w",0])
        self.EventosEscenario.accept("s", self.setKey, ["s",1])
        self.EventosEscenario.accept("s-up", self.setKey, ["s",0])
        self.EventosEscenario.accept("a", self.setKey, ["a",1])
        self.EventosEscenario.accept("a-up", self.setKey, ["a",0])
        self.EventosEscenario.accept("d", self.setKey, ["d",1])
        self.EventosEscenario.accept("d-up", self.setKey, ["d",0])
        
        
        self.EventosEscenario.accept("mouse1", self.setKey, ["mouse1",1])
        self.EventosEscenario.accept("mouse1-up", self.setKey, ["mouse1",0])
        self.EventosEscenario.accept("mouse2", self.setKey, ["mouse2",1])
        self.EventosEscenario.accept("mouse2-up", self.setKey, ["mouse2",0])
        """
       
    def Salir(self):
        
        print 'Escenario:Salir'
        
        # TASK
        taskMgr.remove('actualiza_escenario')
        
        # EVENTOS
        self.EventosEscenario.ignoreAll()
        
        # CIRCULACION
        self.circulacion.Eliminar()
        
        # ELIMINAR TODOS LOS NODEPATH CREADOS
        # borra los elementos del escenario
        while self.Figuras:
            self.Figuras.pop()
        
        # vacia la lista de calles disponibles
        while self.listaCalles:
            self.listaCalles.pop()
        
        while self.listaNodePathRecord:
            self.listaNodePathRecord[0].removeNode()
            self.listaNodePathRecord.pop(0)
        
        
        # MAPA MINI
        self.mapa_mini.Ocultar()
        
        # LIMITES DEL MAPA
        self.cNodePathBoundsMapa.removeNode()        
        
        # DETACH
        
        self.MenuDireccionNodePath.detachNode()

        self.NodoEscenario.detachNode()
        self.NodoEscenario2d.detachNode()

        # MENU

        # JUEGO

        # task
        taskMgr.remove('juego')
        
        # eventos
        self.EventosEscenario.ignore('control-j')

        # remove
        while self.listaNodePathJuego:
            self.listaNodePathJuego[0].removeNode()
            self.listaNodePathJuego.pop(0)
        
        self.textoJuego.detachNode()
        self.textoJuegoTiempo.detachNode()
        
#        RECORD
        
        while self.listaNodePathRecord:
            
            self.listaNodePathRecord[0].removeNode()
            self.listaNodePathRecord.pop(0)
    
        taskMgr.remove('record')

        self.textoRecordTiempo.detachNode()
        self.textoRecord.detachNode()
        self.textoRecordInicio.detachNode()
        
#        RUTA
        # elimina el camino marcado
        self.camino.EliminarCamino()
        
#        # ERROR
        self.textoError.detachNode()    
        
        self.musica.stop()
        
        """
        #cancela los eventos del teclado
        self.EventosEscenario.ignore("w")
        self.EventosEscenario.ignore("w-up")
        self.EventosEscenario.ignore("s")
        self.EventosEscenario.ignore("s-up")
        self.EventosEscenario.ignore("a")
        self.EventosEscenario.ignore("a-up")
        self.EventosEscenario.ignore("d")
        self.EventosEscenario.ignore("d-up")
        self.EventosEscenario.ignore("mouse1")
        self.EventosEscenario.ignore("mouse1-up")
        self.EventosEscenario.ignore("mouse2")
        self.EventosEscenario.ignore("mouse2-up")
        """
    
        # elmiminar la funcione del bucle de ejecucion del videojuego
        #taskMgr.remove(self.Movimiento)
 
    # termina la configuracion de escenario
    def setEscenario(self):
        
        print 'Escenario:setEscenario'
        
        # EVENTOS
        
        self.EventosEscenario.ignore('enter')
        self.EventosEscenario.ignore('arrow_up')
        self.EventosEscenario.ignore('arrow_down')
        self.EventosEscenario.ignore('shift-enter')
        
##        # ELIMINADO CAMBIAR AL DIMENSION DEL MAPA MINI
#        self.EventosEscenario.accept('control-+',self.mapa_mini.AumentarZoom)
#        self.EventosEscenario.accept('control--',self.mapa_mini.DisminuirZoom)
    
        # solo si hay direcciones disponibles se podra buscar una ruta
        if self.listaCalles.__len__()>0:
            
            self.ActivarMenus()
#            self.EventosEscenario.accept('control-f',self.MenuRuta)
#            self.EventosEscenario.accept('control-g',self.MenuIrA)
        
        # ESCENARIO 
        self.ProcesarPorcionMapa()
        
        # LIMITES DEL MAPA
        self.EventosEscenario.accept(self.circulacion.getcNodePathNameJugador()+'-into-mini_bounds',self.LimiteMiniBounds)
        self.EventosEscenario.accept(self.circulacion.getcNodePathNameJugador()+'-into-bounds',self.LimiteBounds)
        
        self.cNodePathBoundsJugador.setPos( self.circulacion.getPosJugador() )       
        
        # TASK 
        self.task_escenario = taskMgr.add(self.ActualizarEscenario,'actualiza_escenario')
        
        
    def ProcesarPorcionMapa(self):
        
        print 'Escenario:ProcesarPorcionMapa'
        
        # Carreteras
        
        for c in self.mapa.getPorcionMapa()['carreteras']:
            self.Figuras.append( Highway(c,self.mapa,self.NodoEscenario) )
        
        for c in self.mapa.getPorcionMapa()['peatonales']:
            self.Figuras.append( Peatonal(c,self.NodoEscenario) )
        
        # Plazas y jardines
        
        for c in self.mapa.getPorcionMapa()['plazas']:
            self.Figuras.append( Plaza(c,self.NodoEscenario) )
            
        for c in self.mapa.getPorcionMapa()['garden']:
            self.Figuras.append( Garden(c,self.NodoEscenario) )
        
        # Iconos
            
        for c in self.mapa.getPorcionMapa()['icono_restaurante']:
            self.Figuras.append( Icono(c,self.NodoEscenario,'restaurante') )
            
        for c in self.mapa.getPorcionMapa()['icono_bar']:
            self.Figuras.append( Icono(c, self.NodoEscenario, 'bar') )
            
        for c in self.mapa.getPorcionMapa()['icono_cafe']:
            self.Figuras.append( Icono(c, self.NodoEscenario, 'cafe') )
            
        for c in self.mapa.getPorcionMapa()['icono_guarderia']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'guarderia') )
            
        for c in self.mapa.getPorcionMapa()['icono_colegio']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'colegio') )
            
        for c in self.mapa.getPorcionMapa()['icono_libreria']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'libreria') )
            
        for c in self.mapa.getPorcionMapa()['icono_universidad']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'universidad') )
            
        for c in self.mapa.getPorcionMapa()['icono_autobus']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'autobus') )
            
        for c in self.mapa.getPorcionMapa()['icono_gasolinera']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'gasolinera') )
            
        for c in self.mapa.getPorcionMapa()['icono_parking']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'parking') )
            
        for c in self.mapa.getPorcionMapa()['icono_taxi']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'taxi') )
            
        for c in self.mapa.getPorcionMapa()['icono_cajero']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'cajero') )
            
        for c in self.mapa.getPorcionMapa()['icono_banco']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'banco') )
            
        for c in self.mapa.getPorcionMapa()['icono_farmacia']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'farmacia') )
            
        for c in self.mapa.getPorcionMapa()['icono_hospital']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'hospital') )
            
        for c in self.mapa.getPorcionMapa()['icono_cine']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'cine') )
            
        for c in self.mapa.getPorcionMapa()['icono_teatro']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'teatro') )
            
        for c in self.mapa.getPorcionMapa()['icono_iglesia']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'iglesia') )
            
        for c in self.mapa.getPorcionMapa()['icono_mezquita']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'mezquita') )
            
        for c in self.mapa.getPorcionMapa()['icono_policia']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'policia') )
            
        for c in self.mapa.getPorcionMapa()['icono_oficina_correos']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'oficina_correos') )

        for c in self.mapa.getPorcionMapa()['icono_museo']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'museo') )
            
        for c in self.mapa.getPorcionMapa()['icono_hotel']:
            self.Figuras.append( Icono(c,self.NodoEscenario, 'hotel') )
            
        # Areas
        
        for c in self.mapa.getPorcionMapa()['area_restaurante']:
            self.Figuras.append( Area(c,self.NodoEscenario,'restaurante') )
            
        for c in self.mapa.getPorcionMapa()['area_bar']:
            self.Figuras.append( Area(c, self.NodoEscenario, 'bar') )
            
        for c in self.mapa.getPorcionMapa()['area_cafe']:
            self.Figuras.append( Area(c, self.NodoEscenario, 'cafe') )
            
        for c in self.mapa.getPorcionMapa()['area_guarderia']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'guarderia') )
            
        for c in self.mapa.getPorcionMapa()['area_colegio']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'colegio') )
            
        for c in self.mapa.getPorcionMapa()['area_libreria']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'libreria') )
            
        for c in self.mapa.getPorcionMapa()['area_universidad']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'universidad') )
            
        for c in self.mapa.getPorcionMapa()['area_autobus']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'autobus') )
            
        for c in self.mapa.getPorcionMapa()['area_gasolinera']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'gasolinera') )
            
        for c in self.mapa.getPorcionMapa()['area_parking']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'parking') )
            
        for c in self.mapa.getPorcionMapa()['area_taxi']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'taxi') )
            
        for c in self.mapa.getPorcionMapa()['area_cajero']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'cajero') )
            
        for c in self.mapa.getPorcionMapa()['area_banco']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'banco') )
            
        for c in self.mapa.getPorcionMapa()['area_farmacia']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'farmacia') )
            
        for c in self.mapa.getPorcionMapa()['area_hospital']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'hospital') )
            
        for c in self.mapa.getPorcionMapa()['area_cine']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'cine') )
            
        for c in self.mapa.getPorcionMapa()['area_teatro']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'teatro') )
            
        for c in self.mapa.getPorcionMapa()['area_iglesia']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'iglesia') )
            
        for c in self.mapa.getPorcionMapa()['area_mezquita']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'mezquita') )
            
        for c in self.mapa.getPorcionMapa()['area_policia']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'policia') )
            
        for c in self.mapa.getPorcionMapa()['area_oficina_correos']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'oficina_correos') )

        for c in self.mapa.getPorcionMapa()['area_museo']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'museo') )
            
        for c in self.mapa.getPorcionMapa()['area_hotel']:
            self.Figuras.append( Area(c,self.NodoEscenario, 'hotel') )
            
        for c in self.mapa.getPorcionMapa()['edificios']:
            self.Figuras.append( Edificio(c,self.NodoEscenario) )    
        
    # no se indica la posicion de inicio del jugador
    #  recibe dos nodos si se procede de un mapa http anterior
    def ComenzarCirculacionAutomatico(self,nodo_origen=None,nodo_destino=None):
        
        print 'Escenario:ComenzarCirculacionAutomatico'
        
#        self.inicioNodePath.detachNode()
        if nodo_origen != None:
#            self.mapa.CalcularPorcionMapa(self.mapa.getNodoPosById(nodo_origen),
#                                      self.tam_mapa)
            self.mapa.CalcularPorcionMapa(nodo_origen.getPos(),
                                      self.tam_mapa)
        
         # CIRCULACION
#        self.circulacion.Crear(self.mapa,nodo_origen,nodo_destino)
        self.circulacion.Crear(self.mapa,nodo_origen.getId(),nodo_destino.getId())
#        self.circulacion.SeleccionarInicio(mapa)
        
        # termina la configuracion dle escenario
        self.setEscenario()
        
#        MOVIDO A setEscenario
#        
#        # ESCENARIO 
#        self.ProcesarPorcionMapa(self.circulacion.getPosJugador(),self.tam_mapa)
#        
#        # LIMITES DEL MAPA
#        self.EventosEscenario.accept(self.circulacion.getcNodePathNameJugador()+'-into-mini_bounds',self.LimiteMiniBounds)
#        self.EventosEscenario.accept(self.circulacion.getcNodePathNameJugador()+'-into-bounds',self.LimiteBounds)
#        
#        self.cNodePathBoundsJugador.setPos( self.circulacion.getPosJugador() )       
#        
#        # TASK 
#        self.task_escenario = taskMgr.add(self.ActualizarEscenario,'task_escenario')
 
    def ComenzarCirculacionManual(self):
        
        print 'Escenario:ComenzarCirculacionManual'
        
        # obtiene un nodo del way de incio del jugador
        nodo_inicio = self.listaCallesBusqueda[self.indiceCalle].getListaRefs()[0]
        
        self.MenuDireccionNodePath.detachNode()
        
#        self.mapa.CalcularPorcionMapa(self.mapa.getNodoPosById(nodo_inicio),
#                                      self.tam_mapa)
        self.mapa.CalcularPorcionMapa(nodo_inicio.getPos(),
                                      self.tam_mapa)
        
        
        # CIRCULACION
#        self.circulacion.Crear(self.mapa,nodo_inicio)
        self.circulacion.Crear(self.mapa,nodo_inicio)
        
        # termina la configuracion del escenario
        self.setEscenario()
        
#         MOVIDO A setEscenario
#        self.circulacion.SeleccionarInicio(mapa)
#       
#        # ESCENARIO 
#        self.ProcesarPorcionMapa(self.circulacion.getPosJugador(),self.tam_mapa)
#        
#        # LIMITES DEL MAPA
#        self.EventosEscenario.accept(self.circulacion.getcNodePathNameJugador()+'-into-mini_bounds',self.LimiteMiniBounds)
#        self.EventosEscenario.accept(self.circulacion.getcNodePathNameJugador()+'-into-bounds',self.LimiteBounds)
#        
#        self.cNodePathBoundsJugador.setPos( self.circulacion.getPosJugador() )
#        
#        # TASK 
#        self.task_escenario = taskMgr.add(self.ActualizarEscenario,'task_escenario')
        
        

    # configura el solido de colision que determina si el vehiculo se sale del mapa
    def LimiteMiniBounds(self,entry):

        print 'Escenario:LimiteMiniBounds'

        # borra los elementos del escenario
        while self.Figuras:
            self.Figuras.pop()
 
        self.mapa.CalcularPorcionMapa(self.circulacion.getPosJugador(),
                                      self.tam_mapa)
 
        # crea los nuevo elementos del escenario
        self.ProcesarPorcionMapa()
       
        # nuevos limites de la parocion de mapa visible
        self.cNodePathBoundsJugador.setPos( self.circulacion.getPosJugador() )
            

               
#        # limite izquierda
#        pos = Point3(self.circulacion.getPosJugador() - Point3(1*self.tam_mapa, 0, 0))
#        self.cNodePathBounds.node().addSolid( CollisionPlane(Plane(Vec3(1, 0, 0), pos)))
#        
#        # limite derecha
#        pos = Point3(self.circulacion.getPosJugador() + Point3(1*self.tam_mapa, 0, 0))
#        self.cNodePathBounds.node().addSolid( CollisionPlane(Plane(Vec3(-1, 0, 0), pos)))
#        
#        # limite abajo
#        pos = Point3(self.circulacion.getPosJugador() - Point3(0, 1*self.tam_mapa, 0))
#        self.cNodePathBounds.node().addSolid(CollisionPlane(Plane(Vec3(0, 1, 0), pos)))
#        
#        # limite arriba
#        pos = Point3(self.circulacion.getPosJugador() + Point3(0, 1*self.tam_mapa, 0))
#        self.cNodePathBounds.node().addSolid( CollisionPlane(Plane(Vec3(0, -1, 0), pos)))

    def LimiteBounds(self,entry):
        
        print 'Escenario:LimiteBounds'
        
        
        
        if self.modoMapa == 'http':
        
#           obtiene las coordenadas de referencia para el nuevo mapa
            nodo_origen = self.circulacion.getNodoOrigenJugador()
            nodo_destino = self.circulacion.getNodoDestinoJugador()
        
            self.Salir()
            self.textoError.node().setText('Cargando el nuevo mapa.')
            self.textoError.reparentTo(self.NodoEscenario2d)
            self.NodoEscenario2d.reparentTo(aspect2d)
            
            lon = self.mapa.getNodoLonById(self.circulacion.getNodoDestinoJugador())
            lat = self.mapa.getNodoLatById(self.circulacion.getNodoDestinoJugador())    
            self.Entrar('http', lat, lon, nodo_origen, nodo_destino)
            
        
#        if self.modoMapa == 'http':
#        
#            # TASK
#            taskMgr.remove('actualiza_escenario')
#            # CIRCULACION
#            self.circulacion.Eliminar()
#        
#        
#            # CREACION DEL NUEVO MAPA
#            lon = self.mapa.getNodoLonById(self.circulacion.getNodoDestinoJugador())
#            lat = self.mapa.getNodoLatById(self.circulacion.getNodoDestinoJugador())    
#            try:
#                self.nuevoMapa = Mapa("http",Escalas['gr-px'],float(lat),float(lon))
#                
#        #        self.EventosEscenario.acceptOnce('cargar_mapa',self.CargarNuevoMapa)
#                self.CargarNuevoMapa()
#            except:
#                print 'Error. No se pudo cargar el nuevo mapa'
#                self.textoError.reparentTo(self.NodoEscenario2d)
#            # si el mapa fue descargado desde la web
        
    def CargarNuevoMapa(self):
        
        print 'Escenario:CargarNuevoMapa'
        
#        # BORRADO DEL MAPA ANTERIOR
#        
#        # TASK
#        taskMgr.remove('actualiza_escenario')
#        
#        # EVENTOS
#        self.EventosEscenario.ignoreAll()
#        
#        # CIRCULACION
#        self.circulacion.Eliminar()
#        
#        # ELIMINAR TODOS LOS NODEPATH CREADOS
#        # borra los elementos del escenario
#        while self.Figuras:
#            self.Figuras.pop()
#        
#        # vacia la lista de calles disponibles
#        while self.listaCalles:
#            self.listaCalles.pop()
#        
#        while self.listaNodePathRecord:
#            self.listaNodePathRecord[0].removeNode()
#            self.listaNodePathRecord.pop(0)
#        
#        # elimina el camino marcado
#        self.camino.EliminarCamino()
#        
#        # MAPA MINI
#        self.mapa_mini.Ocultar()
#        
#        # LIMITES DEL MAPA
#        self.cNodePathBoundsMapa.removeNode()        
#        
#        # DETACH
#        
#        self.MenuDireccionNodePath.detachNode()
#
#        self.NodoEscenario.detachNode()
#        self.NodoEscenario2d.detachNode()
#
#        # MENU
#
#        # JUEGO
#
#        # task
#        taskMgr.remove('juego')
#        
#        # eventos
#        self.EventosEscenario.ignore('control-j')
#
#        # remove
#        while self.listaNodePathJuego:
#            self.listaNodePathJuego[0].removeNode()
#            self.listaNodePathJuego.pop(0)
#        
#        self.textoJuego.detachNode()
#        self.textoJuegoTiempo.detachNode()
#        
##        RECORD
#        
#        while self.listaNodePathRecord:
#            
#            self.listaNodePathRecord[0].removeNode()
#            self.listaNodePathRecord.pop(0)
#    
#        taskMgr.remove('record')
#
#        self.textoRecordTiempo.detachNode()
#        self.textoRecord.detachNode()
#        self.textoRecordInicio.detachNode()
#        
#        # FIN ELIMINAR MAPA ANTERIOR
        
        self.Salir()
        
            
#        # obtiene las coordenadas de referencia para el nuevo mapa
#        nodo_origen = self.circulacion.getNodoOrigenJugador()
#        nodo_destino = self.circulacion.getNodoDestinoJugador()
#          
#        self.mapa = self.nuevoMapa
#        
#
#        
#        # LIMITES DEL MAPA
#        cNodeBounds = CollisionNode( 'bounds' )
#        cNodeBounds.addSolid( CollisionPlane(Plane(Vec3(1, 0, 0), Point3(self.mapa.getBounds().getMinPos()) )) )
#        cNodeBounds.addSolid( CollisionPlane(Plane(Vec3(0, 1, 0), Point3(self.mapa.getBounds().getMinPos()) )) )
#        cNodeBounds.addSolid( CollisionPlane(Plane(Vec3(-1, 0, 0), Point3(self.mapa.getBounds().getMaxPos()) )) )
#        cNodeBounds.addSolid( CollisionPlane(Plane(Vec3(0, -1, 0), Point3(self.mapa.getBounds().getMaxPos()) )) )
#        cNodeBounds.setFromCollideMask(BitMask32(0x22))
#        cNodeBounds.setIntoCollideMask(BitMask32(0x02))
#        self.cNodePathBoundsMapa = self.NodoEscenario.attachNewNode(cNodeBounds)
##        self.cNodePathBounds.show() 
#        
#        # RECOPILACION DE LOS NOMBRE DE LAS CALLES
#        for w_id in self.mapa.getListaWays():
#            
#            w = self.mapa.getWay(w_id)           
#            # si es una carreteras
#            if w.has_tag('highway'):
#                
#                # obtiene su nombre
#                nombre = w.getTagValue('name')
#                if nombre and w.getListaRefs().__len__() > 0:
#                    self.listaCalles.append( (nombre,w.getListaRefs()) )
#       
#       # CIRCULACION
#        self.circulacion.Crear(self.mapa,nodo_origen,nodo_destino)
#        
#        # termina la configuracion del escenario
#        self.setEscenario()
#       
#        self.textoError.detachNode()
            
        """
        1.notificar y salir
        2.intentar descargar un nuevo mapa via http
        """
 
    def ActualizarEscenario(self,task):
        
        # MAPAMINI
        pos = self.circulacion.getPosJugador()
        # altura de la camara, NO es el zoom
        pos.setZ(500)
        self.mapa_mini.setPos( pos )
        # gira la flecha de direccion
        self.mapa_mini.setRoll( self.circulacion.getHeadingJugador() )
        
        
        # CIELO
        self.cielo.setPos(base.camera.getPos())        

        #SUELO
        self.suelo.setPos(base.camera.getPos())
        
        return Task.cont

    def MenuRuta(self):
        
        print 'Escenario:MenuRuta'
        
        # descativa los accesos a menu
        self.DesactivarMenus()
#        self.EventosEscenario.ignore('control-f')
#        self.EventosEscenario.ignore('control-g')
        
        # desactica los eventos de la circualcion
        self.circulacion.DesactivarTeclado()
        # detiene la circulacion del trafico
        self.circulacion.Parar()
        
        # elimina el camino marcado previamente
        self.camino.EliminarCamino()
        
        # borra la caja de texto
        self.DEntryBuscarDir.enterText('')
        # añadde todas las direcciones a la lista, si hay direcciones disponibles lo activa
        self.BuscarDireccionSeleccionada(self.CrearRuta)
        
        self.textoMenuDireccion.node().setText('Dirección a buscar: \n'+
                                                 self.listaCallesBusqueda[self.indiceCalle].getName()+
                                                 '\n('+str(self.indiceCalle+1)+'/'+
                                                 str(self.listaCallesBusqueda.__len__())+
                                                 ')')
            
        # para buscar las calles coincidentes
        self.EventosEscenario.accept('shift-enter',self.BuscarDireccionSeleccionada,[self.CrearRuta])
        # evento para acceder al escenario
#        self.EventosEscenario.accept('enter',self.CrearRuta)
    
        # seleccion de la posicion de inicio
        self.EventosEscenario.accept('arrow_up',self.WayAnterior)
        self.EventosEscenario.accept('arrow_down',self.WaySiguiente)
        
        # evento para salir del menu
        self.EventosEscenario.accept('control-f',self.SalirMenuRuta)
        
        # foco sobre la caja de texto siemre activo
        self.DEntryBuscarDir.enterText('')
        self.DEntryBuscarDir['focus']=1
        
        # muestra el menu
        self.MenuDireccionNodePath.reparentTo(self.NodoEscenario2d)
        
    def CrearRuta(self):
        
        print 'Escenario:CrearRuta'
        
        origen = self.circulacion.getNodoDestinoJugador()
        destino = self.listaCallesBusqueda[self.indiceCalle].getListaRefs()[0]
        
        lista = self.mapa.getRuta(origen,destino)
        if lista:
            self.camino.MostrarCamino(lista)
        else:
            print 'No hay ruta'
            
        self.circulacion.Reanudar()
    
        # Elimina los eventos del menu direccion
        self.EventosEscenario.ignore('enter')
        self.EventosEscenario.ignore('arrow_up')
        self.EventosEscenario.ignore('arrow_down')
        self.EventosEscenario.ignore('shift-enter')
        
        # oculta el menu de direcciones
        self.MenuDireccionNodePath.detachNode()
    
        # reactiva la busqueda de destinos
        self.ActivarMenus()
#        self.EventosEscenario.accept('control-f',self.MenuRuta)
#        self.EventosEscenario.accept('control-g',self.MenuIrA)
        
        # reactiva los eventos de la circulacion
        self.circulacion.ActivarTeclado()
    
    def SalirMenuRuta(self):
        
        print 'Escenario.SalirMenuRuta'
        
        self.circulacion.Reanudar()
    
        # Elimina los eventos del menu direccion
        self.EventosEscenario.ignore('enter')
        self.EventosEscenario.ignore('arrow_up')
        self.EventosEscenario.ignore('arrow_down')
        self.EventosEscenario.ignore('shift-enter')
        self.EventosEscenario.ignore('control-f')
        
        # oculta el menu de direcciones
        self.MenuDireccionNodePath.detachNode()
    
        # reactiva la busqueda de destinos
        self.ActivarMenus()
#        self.EventosEscenario.accept('control-f',self.MenuRuta)
#        self.EventosEscenario.accept('control-g',self.MenuIrA)
        
        # reactiva los eventos de la circulacion
        self.circulacion.ActivarTeclado()

    
    def MenuIrA(self):
        
        print 'Escenario:MenuIrA'
        
        # descativa la entrada a los menu
        self.DesactivarMenus()
#        self.EventosEscenario.ignore('control-f')
#        self.EventosEscenario.ignore('control-g')
        
        # desactica los eventos de la circualcion
        self.circulacion.DesactivarTeclado()
        # detiene la circulacion del trafico
        self.circulacion.Parar()
        
        # borra la caja de texto
        self.DEntryBuscarDir.enterText('')
        # añadde todas las direcciones a la lista, si hay direcciones disponibles lo activa
        self.BuscarDireccionSeleccionada(self.IrA)
        
        self.textoMenuDireccion.node().setText('Dirección a ir: \n'+
                                                 self.listaCallesBusqueda[self.indiceCalle].getName()+
                                                 '\n('+str(self.indiceCalle+1)+'/'+
                                                 str(self.listaCallesBusqueda.__len__())+
                                                 ')')
            
        # para buscar las calles coincidentes
        self.EventosEscenario.accept('shift-enter',self.BuscarDireccionSeleccionada,[self.IrA])
        # evento para acceder al escenario
#        self.EventosEscenario.accept('enter',self.CrearRuta)
    
        # seleccion de la posicion de inicio
        self.EventosEscenario.accept('arrow_up',self.WayAnterior)
        self.EventosEscenario.accept('arrow_down',self.WaySiguiente)
        
        # evento para salir dle menu
        self.EventosEscenario.accept('control-g',self.SalirMenuIrA)
        
        # foco sobre la caja de texto siemre activo
        self.DEntryBuscarDir.enterText('')
        self.DEntryBuscarDir['focus']=1
        
        # muestra el menu
        self.MenuDireccionNodePath.reparentTo(self.NodoEscenario2d)

    """
    Transporta al vehicula a la direccion seleccionada
    """
    def IrA(self):
        
        print 'Escenario:IrA'
        
        self.circulacion.Eliminar()
        
        self.circulacion.Crear(self.mapa, self.listaCallesBusqueda[self.indiceCalle].getListaRefs()[0])

        # oculta el menu de direcciones
        self.MenuDireccionNodePath.detachNode()
    
        # Elimina los eventos del menu 
        self.EventosEscenario.ignore('enter')
        self.EventosEscenario.ignore('arrow_up')
        self.EventosEscenario.ignore('arrow_down')
        self.EventosEscenario.ignore('shift-enter')
        self.EventosEscenario.ignore('control-g')
        
        # reactiva los menus
        self.ActivarMenus()
#        self.EventosEscenario.accept('control-f',self.MenuRuta)
#        self.EventosEscenario.accept('control-g',self.MenuIrA)

        
    
    def SalirMenuIrA(self):
        
        print 'Escenario.SalirMenuIrA'
    
        # oculta el menu de direcciones
        self.MenuDireccionNodePath.detachNode()
    
        # Elimina los eventos del menu 
        self.EventosEscenario.ignore('enter')
        self.EventosEscenario.ignore('arrow_up')
        self.EventosEscenario.ignore('arrow_down')
        self.EventosEscenario.ignore('shift-enter')
        self.EventosEscenario.ignore('control-g')
        
        # reactiva los menus
        self.ActivarMenus()
#        self.EventosEscenario.accept('control-f',self.MenuRuta)
#        self.EventosEscenario.accept('control-g',self.MenuIrA)
        
        self.circulacion.Reanudar()
        
        # reactiva los eventos de la circulacion
        self.circulacion.ActivarTeclado()
        
    """
    Muestra el menu con la lista de direcciones disponibles y la caja de texto a buscar
    Recibe como parámetro la función a ejecutar hecha la seleccion
    """
    def MenuDireccion(self):
        
        # borra la caja de texto
        self.DEntryBuscarDir.enterText('')
        # añadde todas las direcciones a la lista
        self.BuscarDireccionSeleccionada(self.ComenzarCirculacionManual)
        
        self.textoMenuDireccion.node().setText('Dirección de inicio: \n'+
                                                 self.listaCallesBusqueda[self.indiceCalle].getName()+
                                                 '\n('+str(self.indiceCalle+1)+'/'+
                                                 str(self.listaCallesBusqueda.__len__())+
                                                 ')')
            
        self.MenuDireccionNodePath.reparentTo(self.NodoEscenario2d)
        
        # EVENTOS
        
        # Eliminado
        #self.EventosEscenario.accept('c',self.ComenzarCirculacionAutomatico)
        
        
        # para buscar las calles coincidentes
        self.EventosEscenario.accept('shift-enter',self.BuscarDireccionSeleccionada,[self.ComenzarCirculacionManual])
        # evento para acceder al escenario
#        self.EventosEscenario.accept('enter',self.ComenzarCirculacionManual)
    
        # seleccion de la posicion de inicio
        self.EventosEscenario.accept('arrow_up',self.WayAnterior)
        self.EventosEscenario.accept('arrow_down',self.WaySiguiente)
        
        # foco sobre la caja de texto siemre activo
        self.DEntryBuscarDir.enterText('')
        self.DEntryBuscarDir['focus']=1
    
    def MenuRecord(self):
        
        print 'Escenario:MenuRecord'
        
        # descativa la entrada a los menu
        self.DesactivarMenus()
        
        # desactica los eventos de la circualcion
        self.circulacion.DesactivarTeclado()
        # detiene la circulacion del trafico
        self.circulacion.Parar()
        
        # borra la caja de texto
        self.DEntryBuscarDir.enterText('')
        # añadde todas las direcciones a la lista, si hay direcciones disponibles lo activa
        self.BuscarDireccionSeleccionada(self.MedirRecord)
        
        self.textoMenuDireccion.node().setText('Selecciona la meta: \n'+
                                                 self.listaCallesBusqueda[self.indiceCalle].getName()+
                                                 '\n('+str(self.indiceCalle+1)+'/'+
                                                 str(self.listaCallesBusqueda.__len__())+
                                                 ')')
        
        # recupera el id y nobre de la direccion de inicio
        nodo_origen = self.circulacion.getNodoOrigenJugador()
        self.record_way_origen = nodo_origen.getListaWays()[0]
        self.record_way_origen_name = self.record_way_origen.getName()

        if self.record_way_origen_name is None:
            self.record_way_origen_name = 'id: ' + self.record_way_origen.getId()
        
        # muestra la direccion de inicio
        self.textoRecordInicio.node().setText('Direccion de inicio: ' + self.record_way_origen_name)
        self.textoRecordInicio.reparentTo(self.NodoEscenario2d)
        
            
        # para buscar las calles coincidentes
        self.EventosEscenario.accept('shift-enter',self.BuscarDireccionSeleccionada,[self.MedirRecord])
        # evento para acceder al escenario
#        self.EventosEscenario.accept('enter',self.CrearRuta)
    
        # seleccion de la posicion de inicio
        self.EventosEscenario.accept('arrow_up',self.WayAnterior)
        self.EventosEscenario.accept('arrow_down',self.WaySiguiente)
        
        # evento para salir dle menu
        self.EventosEscenario.accept('control-h',self.SalirMenuRecord)
        
        # foco sobre la caja de texto siemre activo
        self.DEntryBuscarDir.enterText('')
        self.DEntryBuscarDir['focus']=1
        
        # muestra el menu
        self.MenuDireccionNodePath.reparentTo(self.NodoEscenario2d)
    
    def MedirRecord(self):
        
        print 'Escenario:MedirRecord'
        
        # obtiene los nodos del way de destino
        refs = self.listaCallesBusqueda[self.indiceCalle].getListaRefs() 
        
         
        # recupera la direccion de destino
        self.record_way_destino = self.listaCallesBusqueda[self.indiceCalle]
        self.record_tiempo_anterior = self.records.RecuperarRecord(self.record_way_origen.getId(),
                                                                    self.record_way_destino.getId()) 
        
        i = 0
        for n in refs:
        
            # obtiene la posicion del nodo
            pos = n.getPos()
            
            # carga la diana  sobre el suelo
            diana = OnscreenImage(image="imagenes/diana.png",parent=self.NodoEscenario)
            diana.setScale(Escalas['m-px']*3)
            diana.setHpr(0,-90,0)
            diana.setPos(pos)
            diana.setTransparency(TransparencyAttrib.MAlpha)
            diana.setColor(1,1,1,0.8)
        
            prioridad = 120
            diana.setBin("fixed", prioridad)
            diana.setDepthTest(False)
      
            self.listaNodePathRecord.append(diana)
            
            # carga el nodo de colision
            
            cNodediana = CollisionNode( 'diana_'+str(i) )
            cNodediana.addSolid( CollisionSphere(0,0,0,Escalas['m-px']*2.5) )
            cNodediana.setFromCollideMask(BitMask32(0x30))
            cNodediana.setIntoCollideMask(BitMask32(0x02))
            cNodePathDiana = NodePath(cNodediana)
            cNodePathDiana.reparentTo(self.NodoEscenario)
            cNodePathDiana.setPos(pos)
            self.listaNodePathRecord.append(cNodePathDiana)
    
            # registra la colision
            self.EventosEscenario.acceptOnce(self.circulacion.getcNodePathNameJugador()+'-into-'+'diana_'+str(i),
                                      self.RecordConseguido)
                
            i = i+1
        
        
        # oculta el menu de direcciones
        self.MenuDireccionNodePath.detachNode()
    
        # oculta la direccion de inicio
        self.textoRecordInicio.detachNode()
    
        # Elimina los eventos del menu 
        self.EventosEscenario.ignore('enter')
        self.EventosEscenario.ignore('arrow_up')
        self.EventosEscenario.ignore('arrow_down')
        self.EventosEscenario.ignore('shift-enter')
        self.EventosEscenario.ignore('control-h')
        
        self.circulacion.Reanudar()
        self.circulacion.ActivarTeclado()
        
#        # incializa el tiempo 
#        self.recordTiempo = 0
        
        # comienza la medicion del tiempo
        taskMgr.add(self.TiempoRecord,'record')
        self.textoRecordTiempo.reparentTo(self.NodoEscenario2d)
    
        self.EventosEscenario.accept('control-h',self.SalirMenuRecord)
    
    def TiempoRecord(self,task):
        
        self.recordTiempo = task.time
             
        if self.record_tiempo_anterior != None:
            self.textoRecordTiempo.node().setText('Tiempo: '+ str(self.recordTiempo).split('.')[0]+
                                            ' seg/'+str(self.record_tiempo_anterior)+' seg')
        else:
            self.textoRecordTiempo.node().setText('Tiempo: '+ str(self.recordTiempo).split('.')[0]+' seg')
        
        return Task.cont
        
    def RecordConseguido(self,entry):
        
        print 'Escenario:RecordConseguido'
        
        # para el tiempo
        taskMgr.remove('record')
        
        # detiene el vehiculo
        self.circulacion.Parar()
        
        # muestra los records
        if self.record_tiempo_anterior == None:
             self.textoRecord.node().setText('Ha estblecido un nuevo record.\n'+
                                             'Tiempo: ' + str(self.recordTiempo)+' seg\n'+
                                             'Origen: '+self.record_way_origen_name+' seg\n'+
                                             'Destino: '+self.listaCallesBusqueda[self.indiceCalle].getName() )
             self.records.GuardarRecord(self.record_way_origen, self.record_way_destino, self.recordTiempo)
        
        elif self.recordTiempo < self.record_tiempo_anterior:
            self.textoRecord.node().setText('Ha estblecido un nuevo record.\n'+
                                             'Tiempo: ' + str(self.recordTiempo)+' seg\n'+
                                             'Tiempo anterior: ' + str(self.record_tiempo_anterior)+' seg\n'+
                                             'Origen: '+self.record_way_origen_name+'\n'+
                                             'Destino: '+self.listaCallesBusqueda[self.indiceCalle].getName() )
            self.records.GuardarRecord(self.record_way_origen, self.record_way_destino, self.recordTiempo)
        
        else:
            self.textoRecord.node().setText('Demasiado tarde.\n'+
                                             'Tiempo: ' + str(self.recordTiempo)+' seg\n'+
                                             'Tiempo anterior: ' + str(self.record_tiempo_anterior)+' seg\n'+
                                             'Origen: '+self.record_way_origen_name+'\n'+
                                             'Destino: '+self.listaCallesBusqueda[self.indiceCalle].getName() )
            
        self.textoRecord.reparentTo(self.NodoEscenario2d)
        
    
    def SalirMenuRecord(self):
    
        print 'Escenario:SalirMenuRecord'
    
        while self.listaNodePathRecord:
            
            self.listaNodePathRecord[0].removeNode()
            self.listaNodePathRecord.pop(0)
    
        taskMgr.remove('record')

        self.textoRecordTiempo.detachNode()
        self.textoRecord.detachNode()
        self.textoRecordInicio.detachNode()

        # oculta el menu de direcciones
        self.MenuDireccionNodePath.detachNode()        
        
        # Elimina los eventos del menu 
        self.EventosEscenario.ignore('enter')
        self.EventosEscenario.ignore('arrow_up')
        self.EventosEscenario.ignore('arrow_down')
        self.EventosEscenario.ignore('shift-enter')
        self.EventosEscenario.ignore('control-h')
        
        # reactiva los menus
        self.ActivarMenus()

        self.circulacion.Reanudar()
        
        # reactiva los eventos de la circulacion
        self.circulacion.ActivarTeclado()
    
    
    def MenuJuego(self):
        
        print 'Escenario:MenuRecord'
        
        # descativa la entrada a los menu
        self.DesactivarMenus()
        
        # desactica los eventos de la circualcion
        self.circulacion.DesactivarTeclado()
        # detiene la circulacion del trafico
        self.circulacion.Parar()
        
        # selecciona una posicion alearia
        self.juegoDestino = randint(0, self.listaCalles.__len__()-1)  
#        self.listaCalles[ self.juegoDestino ]
        
        self.textoJuego.node().setText('Tu hermano pequeño se ha perdido, traelo a casa.'+'\n'
                                     'Está en '+ self.listaCalles[ self.juegoDestino ].getName())
    
        self.textoJuego.reparentTo(self.NodoEscenario2d)
                            
        # evento para salir del menu
        self.EventosEscenario.accept('control-j',self.SalirMenuJuego)
    
        # evento para comenzar el juego
        self.EventosEscenario.accept('enter',self.ComenzarJuego)
    
    def ComenzarJuego(self):
        
        print 'Escenario:ComenzarJuego'
        
        # detach
        self.textoJuego.detachNode()
        
        # eventos
        self.EventosEscenario.ignore('enter')
        
        # carga del avatar
        juegoAvatar = loader.loadModel('modelos/littlebrother')
        juegoAvatar.setPos( self.listaCalles[ self.juegoDestino ].getListaRefs()[0].getPos() )
        juegoAvatar.setScale(Escalas['m-px']*1.5/2)
        
        prioridad = 120
        juegoAvatar.setBin("fixed", prioridad)
#        juegoAvatar.setDepthTest(False)
        juegoAvatar.reparentTo(self.NodoEscenario)
        
        self.listaNodePathJuego.append(juegoAvatar)
        
        # carga el nodo de colision
        cNodeAvatar = CollisionNode('avatar')
        cNodeAvatar.addSolid( CollisionSphere(0,0,0,Escalas['m-px']*2.5) )
        cNodeAvatar.setFromCollideMask(BitMask32(0x30))
        cNodeAvatar.setIntoCollideMask(BitMask32(0x02))
        cNodePathAvatar = NodePath(cNodeAvatar)
        cNodePathAvatar.setPos(self.listaCalles[self.juegoDestino].getListaRefs()[0].getPos() )
        cNodePathAvatar.reparentTo(self.NodoEscenario)
        
        self.listaNodePathJuego.append(cNodePathAvatar)
        
        # flecha de orientacion bajo el coche
        self.flechaJuego = OnscreenImage(image="imagenes/flecha4.png",parent=self.NodoEscenario)
        self.flechaJuego.setScale(Escalas['m-px']*3)
        self.flechaJuego.setHpr(0,-90,0)
        
        self.flechaJuego.setTransparency(TransparencyAttrib.MAlpha)
        self.flechaJuego.setColor(1,1,1,0.8)
        
        prioridad = 110
        self.flechaJuego.setBin("fixed", prioridad)
                               
        self.listaNodePathJuego.append(self.flechaJuego)

        # task
        taskMgr.add(self.JuegoTask,'juego')

        # registra la colision
        self.EventosEscenario.acceptOnce(self.circulacion.getcNodePathNameJugador()+'-into-'+'avatar',
                                  self.JuegoConseguido)
        
        # tiempo
        self.textoJuegoTiempo.reparentTo(self.NodoEscenario2d)
        
        # reanudar
        self.circulacion.Reanudar()
        self.circulacion.ActivarTeclado()

    def JuegoTask(self,task):
        
        # muestra el tiempo transcurrido, y el nombre de la dirección de destino
        self.textoJuegoTiempo.node().setText('Tiempo: ' + str(task.time).split('.')[0]+
                                             ' seg, Direccion: ' + self.listaCalles[ self.juegoDestino ].getName())
        
        # flecha orientada al avatar
        self.flechaJuego.lookAt(self.listaNodePathJuego[0])
        self.flechaJuego.setP(-90)
        self.flechaJuego.setPos(self.circulacion.getPosJugador())
        
        return Task.cont

    
    def JuegoConseguido(self,entry):
        
        print 'Escenario:JuegoConseguido'
        
        taskMgr.remove('juego')
        
        self.textoJuego.node().setText('Enhorabuena, lo encontraste.')
        self.textoJuego.reparentTo(self.NodoEscenario2d)
        
        self.circulacion.Parar()
        self.circulacion.DesactivarTeclado()
    
    def SalirMenuJuego(self):
        
        print 'Escenario:SalirJuego'
        
        # task
        taskMgr.remove('juego')
        
        # eventos
        self.EventosEscenario.ignore('control-j')

        # remove
        while self.listaNodePathJuego:
            self.listaNodePathJuego[0].removeNode()
            self.listaNodePathJuego.pop(0)
        
        self.textoJuego.detachNode()
        self.textoJuegoTiempo.detachNode()
        
        # reactiva los menus
        self.ActivarMenus()

        # reanuda la circulacion
        self.circulacion.Reanudar()
        # reactiva los eventos de la circulacion
        self.circulacion.ActivarTeclado()
    
    """
    Desactiva los posibles menus del escenario
    """
    def DesactivarMenus(self):
        
        # menu para buscar una direccion
        self.EventosEscenario.ignore('control-f')
        # menu para ir a una direccion
        self.EventosEscenario.ignore('control-g')
        # menu para medir el tiempo de llegada
        self.EventosEscenario.ignore('control-h')
        # menu para que el jugador busque una direccion
        self.EventosEscenario.ignore('control-j')
    
    def ActivarMenus(self):
        
        self.EventosEscenario.accept('control-f',self.MenuRuta)
        self.EventosEscenario.accept('control-g',self.MenuIrA)
        self.EventosEscenario.accept('control-h',self.MenuRecord)
        self.EventosEscenario.accept('control-j',self.MenuJuego)
    
    # localiza una direccion de la lista de posible segun el texto introducido
    def BuscarDireccionSeleccionada(self,f=None):
        
        print 'Escenario:BuscarDireccionSeleccionada'
        
        # por defecto la primera direccion en aparecer
        self.indiceCalle = 0
        
        # inicialmente no hay direcciones disponibles
        self.listaCallesBusqueda = []
        
        # durante la busqueda desactiva los eventos
        self.EventosEscenario.ignore('arrow_up')
        self.EventosEscenario.ignore('arrow_down')
        self.EventosEscenario.ignore('enter')
        
        # busca los nombre coincidentes
        for c in self.listaCalles:
            if match("(.)*"+self.DEntryBuscarDir.get()+"(.)*",c.getTagValue('name'),IGNORECASE):
                self.listaCallesBusqueda.append(c)
    
        # muestra las direcciones dispoknibles
        if self.listaCallesBusqueda.__len__() > 0: 
            self.textoMenuDireccion.node().setText('Dirección de inicio: \n'+
                                         self.listaCallesBusqueda[self.indiceCalle].getName()+
                                         '\n('+str(self.indiceCalle+1)+
                                         '/'+str(self.listaCallesBusqueda.__len__())+
                                         ')')
            
#            # si hay coincidencias activa de nuevo los eventos
#            self.EventosEscenario.accept('enter',self.ComenzarCirculacionManual)
            self.EventosEscenario.accept('enter',f)
            self.EventosEscenario.accept('arrow_up',self.WayAnterior)
            self.EventosEscenario.accept('arrow_down',self.WaySiguiente)
            
        else:
            # no hay direccione disponibles
            self.textoMenuDireccion.node().setText('Escenario: no hay direcciones disponibles.')
            print 'Escenario: no hay direcciones disponibles.'
            
        # foco activo sobre la caja de texto de busqueda
        self.DEntryBuscarDir['focus']=1
    
    def setFocus(self,arg):
        # foco activo sobre la caja de texto de busqueda
        self.DEntryBuscarDir['focus']=1
        
    
    def WaySiguiente(self):
        
#        print 'Escenario:WaySiguiente'
        
        self.indiceCalle = (self.indiceCalle+1)%self.listaCallesBusqueda.__len__()
        self.textoMenuDireccion.node().setText('Dirección de inicio: \n'+
                                                 self.listaCallesBusqueda[self.indiceCalle].getName()+
                                                 '\n('+str(self.indiceCalle+1)+
                                                 '/'+str(self.listaCallesBusqueda.__len__())+
                                                 ')')
        
    def WayAnterior(self):
        
#        print 'Escenario:WayAnterior'
        
        self.indiceCalle = (self.indiceCalle-1)%self.listaCallesBusqueda.__len__()
        self.textoMenuDireccion.node().setText('Dirección de inicio: \n'+
                                                 self.listaCallesBusqueda[self.indiceCalle].getName()+
                                                 '\n('+str(self.indiceCalle+1)+
                                                 '/'+str(self.listaCallesBusqueda.__len__())+
                                                 ')')
        
    def MostrarControles(self):
        self.textoControles.reparentTo(self.NodoEscenario2d)
    def OcultarControles(self):
        self.textoControles.detachNode()
        

